#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os

import pytorch_lightning as pl
import torch
from pytorch_lightning import Trainer
from torch.nn import functional as F

from zenml.steps.trainer import TorchFeedForwardTrainer
from zenml.steps.trainer import utils
from zenml.utils import path_utils


class MyPyTorchLightningTrainer(TorchFeedForwardTrainer):
    """
    PyTorch Lightning trainer
    """

    def run_fn(self):
        train_split_patterns = [self.input_patterns[split] for split in
                                self.split_mapping[utils.TRAIN_SPLITS]]
        train_dataset = self.input_fn(train_split_patterns)

        eval_split_patterns = [self.input_patterns[split] for split in
                               self.split_mapping[utils.EVAL_SPLITS]]
        eval_dataset = self.input_fn(eval_split_patterns)

        class LitModel(pl.LightningModule):
            def __init__(self):
                super().__init__()
                self.l1 = torch.nn.Linear(8, 64)
                self.layer_out = torch.nn.Linear(64, 1)

            def forward(self, x):
                x = torch.relu(self.l1(x))
                x = self.layer_out(x)
                return x

            def training_step(self, batch, batch_idx):
                x, y, z = batch
                x_batch = torch.cat(list(x.values()), dim=-1)
                y_t = torch.cat(list(y.values()), dim=-1)
                y_hat = self(x_batch)
                loss = F.binary_cross_entropy_with_logits(y_hat, y_t)
                tensorboard_logs = {'train_loss': loss}
                return {'loss': loss, 'log': tensorboard_logs}

            def configure_optimizers(self):
                return torch.optim.Adam(self.parameters(), lr=0.001)

            def train_dataloader(self):
                return train_dataset

            def validation_step(self, batch, batch_idx):
                x, y, z = batch
                x_batch = torch.cat(list(x.values()), dim=-1)
                y_t = torch.cat(list(y.values()), dim=-1)
                y_hat = self(x_batch)
                return {
                    'val_loss': F.binary_cross_entropy_with_logits(y_hat, y_t)}

            def validation_epoch_end(self, outputs):
                avg_loss = torch.stack([x['val_loss'] for x in outputs]).mean()
                tensorboard_logs = {'val_loss': avg_loss}
                return {'avg_val_loss': avg_loss, 'log': tensorboard_logs}

            def val_dataloader(self):
                return eval_dataset

        model = LitModel()

        # most basic trainer, uses good defaults
        trainer = Trainer(
            default_root_dir=self.log_dir,
            max_epochs=self.epochs,
        )
        trainer.fit(model)

        path_utils.create_dir_if_not_exists(self.serving_model_dir)
        if path_utils.is_remote(self.serving_model_dir):
            temp_model_dir = '__temp_model_dir__'
            temp_path = os.path.join(os.getcwd(), temp_model_dir)
            if path_utils.is_dir(temp_path):
                raise PermissionError('{} is used as a temp path but it '
                                      'already exists. Please remove it to '
                                      'continue.')
            trainer.save_checkpoint(os.path.join(temp_path, 'model.cpkt'))
            path_utils.copy_dir(temp_path, self.serving_model_dir)
            path_utils.rm_dir(temp_path)
        else:
            trainer.save_checkpoint(
                os.path.join(self.serving_model_dir, 'model.ckpt'))

        # test
        for split in self.split_mapping[utils.TEST_SPLITS]:
            pattern = self.input_patterns[split]
            test_dataset = self.input_fn([pattern])
            test_results = self.test_fn(model, test_dataset)
            utils.save_test_results(test_results, self.output_patterns[split])



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Text

from tfx.orchestration import metadata

from zenml.metadata import ZenMLMetadataStore
from zenml.enums import MLMetadataTypes


class MySQLMetadataStore(ZenMLMetadataStore):
    STORE_TYPE = MLMetadataTypes.mysql.name

    def __init__(self, host: Text, port: int, database: Text, username: Text,
                 password: Text):
        """Constructor for MySQL MetadataStore for ZenML"""
        self.host = host
        self.port = int(port)  # even if its a string, it should be casted
        self.database = database
        self.username = username
        self.password = password
        super().__init__()

    def get_tfx_metadata_config(self):
        return metadata.mysql_metadata_connection_config(
            host=self.host,
            port=self.port,
            database=self.database,
            username=self.username,
            password=self.password)



from typing import List

import pandas as pd
import panel as pn
import param
import plotly.express as px
import plotly.graph_objects as go
import tensorflow_model_analysis as tfma

from zenml.enums import PipelineStatusTypes, GDPComponent
from zenml.pipelines import TrainingPipeline
from zenml.repo import Repository

pn.extension('plotly')


def parse_metrics(d):
    tmp = d.copy()
    for x in d:
        if x.startswith('metric_'):
            m = tmp.pop(x)
            for k, v in m.items():
                tmp[x + '_' + k] = v['doubleValue']
    return tmp


class Application(param.Parameterized):
    pipeline_run_selector = param.ListSelector(default=[], objects=[])
    hyperparameter_selector = param.ListSelector(default=[], objects=[])

    slicing_metric_selector = param.ObjectSelector(default='', objects=[''])
    performance_metric_selector = param.ObjectSelector(objects=[])

    def __init__(self, datasource=None, **params):
        super(Application, self).__init__(**params)

        # lists
        result_list = []
        hparam_list = []
        repo: Repository = Repository.get_instance()
        self.datasource = datasource

        # get all pipelines in this workspace
        if datasource:
            # filter pipeline by datasource, and then the training ones
            all_pipelines: List[TrainingPipeline] = \
                repo.get_pipelines_by_datasource(datasource)
            all_pipelines = [p for p in all_pipelines if
                             p.PIPELINE_TYPE == TrainingPipeline.PIPELINE_TYPE]
        else:
            all_pipelines: List[TrainingPipeline] = repo.get_pipelines_by_type(
                [
                    TrainingPipeline.PIPELINE_TYPE])

        # get a dataframe of all results + all hyperparameter combinations
        for p in all_pipelines:
            # This is slowing the comparison down but
            # necessary to update the status of each run
            if p.get_status() == PipelineStatusTypes.Succeeded.name:
                eval_path = p.get_artifacts_uri_by_component(
                    GDPComponent.Evaluator.name)[0]

                evaluation = tfma.load_eval_result(eval_path)
                for s, m in evaluation.slicing_metrics:
                    result_list.append(dict([(
                        'pipeline_name', '{}'.format(p.name)),
                        ('slice_name', s[0][0] if s else ''),
                        ('slice_value', s[0][1] if s else '')]))
                    result_list[-1].update({f'metric_{k}': m[k]['']
                                            for k, v in m.items()})

                h_dict = p.get_hyperparameters()
                h_dict['pipeline_name'] = p.name
                hparam_list.append(h_dict)

        self.results = pd.DataFrame([parse_metrics(r) for r in
                                     result_list])
        self.hparam_info = pd.DataFrame(hparam_list)

        # set params
        self.param.pipeline_run_selector.objects = self.results[
            'pipeline_name'].unique()

    @param.depends('pipeline_run_selector', watch=True)
    def _updated_context(self):
        df = self.results[self.results['pipeline_name'].isin(
            self.pipeline_run_selector)]
        df = df.dropna(axis=1, how='all')

        slicing_metric_list = sorted(list(df['slice_name'].unique()))

        performance_metric_set = {c for c in df.columns if
                                  c.startswith('metric_')}
        performance_metric_list = [None] + sorted(list(
            performance_metric_set))

        self.param['slicing_metric_selector'].objects = slicing_metric_list
        self.param[
            'performance_metric_selector'].objects = performance_metric_list

        # get params
        parameter_list = self.hparam_info[self.hparam_info[
            'pipeline_name'].isin(self.pipeline_run_selector)].columns
        parameter_list = [x for x in parameter_list if x !=
                          'pipeline_name']
        self.param['hyperparameter_selector'].objects = parameter_list

        self.slicing_metric_selector = ''
        self.performance_metric_selector = None

    @param.depends('slicing_metric_selector',
                   'performance_metric_selector', watch=True)
    def performance_graph(self):
        if self.performance_metric_selector:
            df = self.results[(self.results['pipeline_name'].isin(
                self.pipeline_run_selector)) &
                              (self.results['slice_name'] ==
                               self.slicing_metric_selector)]
            fig = px.scatter(df,
                             x='pipeline_name',
                             y=self.performance_metric_selector,
                             color='slice_value',
                             width=1100,
                             title='Pipeline Comparison')

            fig = fig.update_traces(mode='lines+markers')

        else:
            fig = px.scatter(pd.DataFrame(),
                             marginal_y='rug',
                             width=1100,
                             title='Pipeline Comparison')

        return fig

    @param.depends('performance_metric_selector',
                   'hyperparameter_selector', watch=True)
    def parameter_graph(self):
        if self.performance_metric_selector and len(
                self.hyperparameter_selector) > 0:
            df = self.results[(self.results['pipeline_name'].isin(
                self.pipeline_run_selector)) &
                              (self.results['slice_name'] == '')]

            # merge
            extra_df = pd.merge(self.hparam_info, df,
                                on='pipeline_name', how='left')

            dimensions = ['pipeline_name'] + self.hyperparameter_selector + \
                         [self.performance_metric_selector]

            new_dims = []
            for d in dimensions:
                try:
                    new_dims.append({'label': d, 'values':
                        pd.to_numeric(extra_df[d])})
                except:
                    u = sorted(list(extra_df[d].apply(lambda x: str(
                        x)).unique()))
                    mapping = {v: i for i, v in enumerate(u)}
                    new_dims.append({'label': d,
                                     'tickvals': [mapping[x] for x
                                                  in u],
                                     'ticktext': u,
                                     'values': extra_df[d].apply(
                                         lambda x: str(x)).map(mapping)})

            final_col = pd.to_numeric(extra_df[
                                          self.performance_metric_selector])
            fig = go.Figure(data=go.Parcoords(line=dict(
                color=final_col,
                colorscale=
                'inferno',
                showscale=True,
                cmin=min(
                    final_col),
                cmax=max(
                    final_col)),
                dimensions=new_dims))
        else:
            fig = px.scatter(pd.DataFrame(),
                             marginal_y='rug',
                             width=1100,
                             title='Hyperparameter Comparison')

        return fig


def generate_interface(datasource=None):
    app = Application(datasource=datasource)
    handlers = pn.Param(app.param)

    # Analysis Page
    analysis_page = pn.GridSpec(height=850, width=1850, max_height
    =850, max_width=1850)
    analysis_page[0:8, 0:2] = handlers[1]
    analysis_page[0:10, 8:10] = handlers[2]
    analysis_page[8:9, 0:2] = handlers[3]
    analysis_page[9:10, 0:2] = handlers[4]
    analysis_page[0:5, 2:8] = app.performance_graph
    analysis_page[5:10, 2:8] = app.parameter_graph

    interface = pn.Tabs(
        ('Analysis Page', analysis_page),
    )
    return interface



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""TFDS Datasource definition"""

import os
from typing import Text, Callable

import tensorflow_datasets as tfds

from zenml.datasources import BaseDatasource
from zenml.utils import path_utils


class TFDSDatasource(BaseDatasource):
    """ZenML TFDS datasource definition."""

    def __init__(
            self,
            name: Text,
            tfds_dataset_name: Text,
            shuffle_files: bool = False,
            as_supervised: bool = False,
            **kwargs):
        """
        ZenML [TFDS](https://www.tensorflow.org/datasets) wrapper.

        Args:
            name: Name of the datasource.
            tfds_dataset_name: Name of the TFDS dataset name, should
            correspond to the `name` argument passed to the `tfds.load` method.
            shuffle_files: Whether to shuffled files or not.
            as_supervised: Corresponds to the `as_supervised` argument of
            the `tfds.load` method.
            **kwargs:
        """
        self.name = name
        self.tfds_dataset_name = tfds_dataset_name
        self.shuffle_files = shuffle_files
        self.as_supervised = as_supervised
        super().__init__(
            name,
            tfds_dataset_name=tfds_dataset_name,
            shuffle_files=shuffle_files,
            as_supervised=as_supervised,
            **kwargs)

    def process(self, output_path: Text, make_beam_pipeline: Callable = None):
        tfds.load(
            self.tfds_dataset_name,
            download=True,
            data_dir=output_path,
            shuffle_files=self.shuffle_files,
            as_supervised=self.as_supervised,
            with_info=True,
        )
        data_dir = os.path.join(output_path, self.tfds_dataset_name)

        for filename in path_utils.find_files(data_dir, '*.tfrecord*'):
            path_utils.copy(filename, os.path.join(output_path,
                                                   os.path.basename(filename)))



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base ZenML repository"""

import os
from pathlib import Path
from typing import Text, List, Dict, Any, Optional, Union, Type

from zenml.exceptions import InitializationException
from zenml.logger import get_logger
from zenml.metadata import ZenMLMetadataStore
from zenml.repo import ArtifactStore, GitWrapper, GlobalConfig, ZenMLConfig
from zenml.repo.constants import ZENML_DIR_NAME
from zenml.standards import standard_keys as keys
from zenml.utils import path_utils, yaml_utils
from zenml.utils.analytics_utils import track, CREATE_REPO, GET_PIPELINES, \
    GET_DATASOURCES, GET_STEPS_VERSIONS, \
    REGISTER_PIPELINE, GET_STEP_VERSION

logger = get_logger(__name__)


class Repository:
    """ZenML repository definition. This is a Singleton class.

    Every ZenML project exists inside a ZenML repository.
    """
    __instance__ = None

    def __init__(self, path: Text = None):
        """
        Construct reference a ZenML repository

        Args:
            path (str): Path to root of repository
        """
        if Repository.__instance__ is None:
            if path is None:
                try:
                    # Start from cwd and traverse up until find zenml config.
                    path = Repository.get_zenml_dir(os.getcwd())
                except Exception:
                    # If there isnt a zenml.config, use the cwd
                    path = os.getcwd()

            if not path_utils.is_dir(path):
                raise Exception(f'{path} does not exist or is not a dir!')
            self.path = path

            # Hook up git, path needs to have a git folder.
            self.git_wrapper = GitWrapper(self.path)

            # Load the ZenML config
            try:
                self.zenml_config = ZenMLConfig(self.path)
            except InitializationException:
                # We allow this because we of the GCP orchestrator for now
                self.zenml_config = None

            Repository.__instance__ = self
        else:
            raise Exception("You cannot create another Repository class!")

    @staticmethod
    def get_zenml_dir(path: Text):
        """
        Recursive function to find the zenml config starting from path.

        Args:
            path (str): starting path
        """
        if ZenMLConfig.is_zenml_dir(path):
            return path

        if path_utils.is_root(path):
            raise Exception(
                'Looks like you used ZenML outside of a ZenML repo. '
                'Please init a ZenML repo first before you using '
                'the framework.')
        return Repository.get_zenml_dir(str(Path(path).parent))

    @staticmethod
    def get_instance(path: Text = None):
        """ Static method to fetch the current instance."""
        logger.debug('Repository instance fetched.')
        if not Repository.__instance__:
            Repository(path)
        return Repository.__instance__

    @staticmethod
    @track(event=CREATE_REPO)
    def init_repo(repo_path: Text, artifact_store_path: Text = None,
                  metadata_store: Optional[ZenMLMetadataStore] = None,
                  pipelines_dir: Text = None,
                  analytics_opt_in: bool = None):
        """
        Initializes a git repo with zenml.

        Args:
            repo_path (str): path to root of a git repo
            metadata_store: metadata store definition
            artifact_store_path (str): path where to store artifacts
            pipelines_dir (str): path where to store pipeline configs.
            analytics_opt_in (str): opt-in flag for analytics code.

        Raises:
            InvalidGitRepositoryError: If repository is not a git repository.
            NoSuchPathError: If the repo_path does not exist.
        """
        # check whether its a git repo by initializing GitWrapper
        git_wrapper = GitWrapper(repo_path)
        # TODO [LOW] Do proper checks and add to .gitignore
        git_wrapper.add_gitignore([ZENML_DIR_NAME + '/'])

        # use the underlying ZenMLConfig class to create the config
        ZenMLConfig.to_config(
            repo_path, artifact_store_path, metadata_store, pipelines_dir)

        # create global config
        global_config = GlobalConfig.get_instance()
        if analytics_opt_in is not None:
            global_config.set_analytics_opt_in(analytics_opt_in)

    def get_default_artifact_store(self) -> Optional[ArtifactStore]:
        self._check_if_initialized()
        return self.zenml_config.get_artifact_store()

    def get_default_metadata_store(self):
        self._check_if_initialized()
        return self.zenml_config.get_metadata_store()

    def get_default_pipelines_dir(self) -> Text:
        self._check_if_initialized()
        return self.zenml_config.get_pipelines_dir()

    def get_git_wrapper(self) -> GitWrapper:
        self._check_if_initialized()
        return self.git_wrapper

    # TODO [MEDIUM]: Potentially move these three to relevant Base classes.
    def get_artifact_store_from_file_path(self, file_path: Text):
        """
        Gets artifact store from a pipeline config file path.
        Args:
            file_path (Text): Path to pipeline YAML file.
        """
        config = yaml_utils.read_yaml(file_path)
        return ArtifactStore(config[keys.GlobalKeys.ARTIFACT_STORE])

    def get_metadata_store_from_file_path(self, file_path: Text):
        """
        Gets metadata store from a pipeline config file path.
        Args:
            file_path (Text): Path to pipeline YAML file.
        """
        config = yaml_utils.read_yaml(file_path)
        return ZenMLMetadataStore.from_config(
            config[keys.GlobalKeys.METADATA_STORE])

    def get_orchestrator_backend_from_file_path(self, file_path: Text):
        """
        Gets orchestrator backend from a pipeline config file path.
        Args:
            file_path (Text): Path to pipeline YAML file.
        """
        from zenml.backends.orchestrator.base.orchestrator_base_backend \
            import OrchestratorBaseBackend
        config = yaml_utils.read_yaml(file_path)
        return OrchestratorBaseBackend.from_config(
            config[keys.GlobalKeys.BACKEND])

    @track(event=GET_STEP_VERSION)
    def get_step_by_version(self, step_type: Union[Type, Text], version: Text):
        """
        Gets a Step object by version. There might be many objects of a
        particular Step registered in many pipelines. This function just
        returns the first configuration that it matches.

        Args:
            step_type: either a string specifying full source of the step or a
            python class type.
            version: either sha pin or standard ZenML version pin.
        """
        from zenml.utils import source_utils
        from zenml.steps.base_step import BaseStep

        type_str = source_utils.get_module_source_from_class(step_type)

        for file_path in self.get_pipeline_file_paths():
            c = yaml_utils.read_yaml(file_path)
            for step_name, step_config in c[keys.GlobalKeys.PIPELINE][
                keys.PipelineKeys.STEPS].items():
                # Get version from source
                class_ = source_utils.get_class_source_from_source(
                    step_config[keys.StepKeys.SOURCE])
                source_version = source_utils.get_pin_from_source(
                    step_config[keys.StepKeys.SOURCE])

                if class_ == type_str and version == source_version:
                    return BaseStep.from_config(step_config)

    def get_step_versions_by_type(self, step_type: Union[Type, Text]):
        """
        List all registered steps in repository by step_type.

        Args:
            step_type: either a string specifying full source of the step or a
            python class type.
        """
        from zenml.utils import source_utils
        type_str = source_utils.get_module_source_from_class(step_type)

        steps_dict = self.get_step_versions()
        if type_str not in steps_dict:
            logger.warning(f'Type {type_str} not available. Available types: '
                           f'{list(steps_dict.keys())}')
            return
        return steps_dict[type_str]

    @track(event=GET_STEPS_VERSIONS)
    def get_step_versions(self):
        """List all registered steps in repository"""
        from zenml.utils import source_utils
        steps_dict = {}
        for file_path in self.get_pipeline_file_paths():
            c = yaml_utils.read_yaml(file_path)
            for step_name, step_config in c[keys.GlobalKeys.PIPELINE][
                keys.PipelineKeys.STEPS].items():
                # Get version from source
                version = source_utils.get_pin_from_source(
                    step_config[keys.StepKeys.SOURCE])
                class_ = source_utils.get_class_source_from_source(
                    step_config[keys.StepKeys.SOURCE])

                # Add to set of versions
                if class_ in steps_dict:
                    steps_dict[class_].add(version)
                else:
                    steps_dict[class_] = {version}
        return steps_dict

    def get_datasource_by_name(self, name: Text):
        """
        Get all datasources in this repo.

        Returns: list of datasources used in this repo
        """
        all_datasources = self.get_datasources()
        for d in all_datasources:
            if name == d.name:
                return d

    def get_datasource_id_by_name(self, name: Text) -> List:
        """
        Get ID of a datasource by just its name.

        Returns: ID of datasource.
        """
        for file_path in self.get_pipeline_file_paths():
            c = yaml_utils.read_yaml(file_path)
            src = c[keys.GlobalKeys.PIPELINE][keys.PipelineKeys.DATASOURCE]
            if keys.DatasourceKeys.NAME in src:
                if name == src[keys.DatasourceKeys.NAME]:
                    return src[keys.DatasourceKeys.ID]

    def get_datasource_names(self) -> List:
        """
        Get all datasources in this repo.

        Returns: List of datasource names used in this repo.
        """
        n = []
        for file_path in self.get_pipeline_file_paths():
            c = yaml_utils.read_yaml(file_path)
            src = c[keys.GlobalKeys.PIPELINE][keys.PipelineKeys.DATASOURCE]
            if keys.DatasourceKeys.NAME in src:
                n.append(src[keys.DatasourceKeys.NAME])
        return list(set(n))

    @track(event=GET_DATASOURCES)
    def get_datasources(self) -> List:
        """
        Get all datasources in this repo.

        Returns: list of datasources used in this repo
        """
        from zenml.datasources import BaseDatasource

        datasources = []
        datasources_name = set()
        for file_path in self.get_pipeline_file_paths():
            c = yaml_utils.read_yaml(file_path)
            ds = BaseDatasource.from_config(c[keys.GlobalKeys.PIPELINE])
            if ds and ds.name not in datasources_name:
                datasources.append(ds)
                datasources_name.add(ds.name)
        return datasources

    def get_pipeline_by_name(self, pipeline_name: Text = None):
        """
        Loads a pipeline just by its name.

        Args:
            pipeline_name (str): Name of pipeline.
        """
        from zenml.pipelines import BasePipeline
        yamls = self.get_pipeline_file_paths()
        for y in yamls:
            n = BasePipeline.get_name_from_pipeline_name(os.path.basename(y))
            if n == pipeline_name:
                c = yaml_utils.read_yaml(y)
                return BasePipeline.from_config(c)

    def get_pipelines_by_type(self, type_filter: List[Text]) -> List:
        """
        Gets list of pipelines filtered by type.

        Args:
            type_filter (list): list of types to filter by.
        """
        pipelines = self.get_pipelines()
        return [p for p in pipelines if p.PIPELINE_TYPE in type_filter]

    def get_pipeline_names(self) -> Optional[List[Text]]:
        """Gets list of pipeline (unique) names"""
        from zenml.pipelines import BasePipeline
        yamls = self.get_pipeline_file_paths(only_file_names=True)
        return [BasePipeline.get_name_from_pipeline_name(p) for p in yamls]

    def get_pipeline_file_paths(self, only_file_names: bool = False) -> \
            Optional[List[Text]]:
        """Gets list of pipeline file path"""
        self._check_if_initialized()

        pipelines_dir = self.zenml_config.get_pipelines_dir()

        if not path_utils.is_dir(pipelines_dir):
            return []
        all_files = path_utils.list_dir(pipelines_dir, only_file_names)
        return [x for x in all_files if yaml_utils.is_yaml(x)]

    def get_pipelines_by_datasource(self, datasource):
        """
        Gets list of pipelines associated with datasource.

        Args:
            datasource (BaseDatasource): object of type BaseDatasource.
        """
        from zenml.pipelines import BasePipeline
        pipelines = []
        for file_path in self.get_pipeline_file_paths():
            c = yaml_utils.read_yaml(file_path)
            if keys.DatasourceKeys.ID in c[keys.GlobalKeys.PIPELINE][
                keys.PipelineKeys.DATASOURCE]:
                if c[keys.GlobalKeys.PIPELINE][keys.PipelineKeys.DATASOURCE][
                    keys.DatasourceKeys.ID] == datasource._id:
                    pipelines.append(BasePipeline.from_config(c))
        return pipelines

    @track(event=GET_PIPELINES)
    def get_pipelines(self) -> List:
        """Gets list of all pipelines."""
        from zenml.pipelines import BasePipeline
        pipelines = []
        for file_path in self.get_pipeline_file_paths():
            c = yaml_utils.read_yaml(file_path)
            pipelines.append(BasePipeline.from_config(c))
        return pipelines

    @track(event=REGISTER_PIPELINE)
    def register_pipeline(self, file_name: Text, config: Dict[Text, Any]):
        """
        Registers a pipeline in the artifact store as a YAML file.

        Args:
            file_name (str): file name of pipeline
            config (dict): dict representation of ZenML config
        """
        self._check_if_initialized()

        pipelines_dir = self.zenml_config.get_pipelines_dir()

        # Create dir
        path_utils.create_dir_if_not_exists(pipelines_dir)

        # Write
        yaml_utils.write_yaml(os.path.join(pipelines_dir, file_name), config)

    def load_pipeline_config(self, file_name: Text) -> Dict[Text, Any]:
        """
        Loads a ZenML config from YAML.

        Args:
            file_name (str): file name of pipeline
        """
        self._check_if_initialized()
        pipelines_dir = self.zenml_config.get_pipelines_dir()
        return yaml_utils.read_yaml(os.path.join(pipelines_dir, file_name))

    def compare_training_runs(self, port: int = 0, datasource=None):
        """Launch the compare app for all training pipelines in repo"""
        from zenml.utils.post_training.post_training_utils import \
            launch_compare_tool
        launch_compare_tool(port, datasource)

    def clean(self):
        """Deletes associated metadata store, pipelines dir and artifacts"""
        raise NotImplementedError

    def _check_if_initialized(self):
        if self.zenml_config is None:
            raise InitializationException



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of the data Processing Backend"""

from typing import Optional, List, Text

from zenml.backends import BaseBackend


class ProcessingBaseBackend(BaseBackend):
    """
    Use this class to run a ZenML pipeline locally.

    Every ZenML pipeline runs in backends.

    A dedicated processing backend can be used to efficiently process large
    amounts of incoming data in parallel, potentially distributed across
    multiple machines. This can happen on local processing backends as well
    as cloud-based variants like Google Cloud Dataflow. More powerful machines
    with higher core counts and clock speeds can be leveraged to increase
    processing throughput significantly.
    """
    BACKEND_TYPE = 'processing'

    def get_beam_args(self,
                      pipeline_name: Text = None,
                      pipeline_root: Text = None) -> \
            Optional[List[Text]]:
        """
        Returns a list of beam args for the pipeline.

        Args:
            pipeline_name: Name of the pipeline.
            pipeline_root: Root dir of pipeline.
        """
        # TODO: [LOW] Check if multiprocessing slows pipeline down or not.
        return [
            # '--direct_running_mode=multi_processing',
            # # 0 means auto-detect based on on the number of CPUs available
            # # during execution time.
            # '--direct_num_workers=0',
        ]



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Type

from tfx.dsl.components.base import executor_spec
from tfx.dsl.components.base.base_executor import BaseExecutor

from zenml.steps import BaseStep


class BaseDeployerStep(BaseStep):
    """
    Base class for all deployer steps. These steps are used to specify a
    configuration for a Pusher component that uploads the model to a selected
    location and opens an endpoint to it from where it serves predictions upon
    request.
    """

    def __init__(self, **kwargs):
        """
        Base deployer step constructor. In order to create custom model
        serving logic, implement deployer steps that inherit from this class.

        Args:
            **kwargs: Additional keyword arguments.
        """
        super(BaseDeployerStep, self).__init__(**kwargs)

    def get_config(self):
        pass

    def build_push_destination(self):
        pass

    def get_executor(self) -> Type[BaseExecutor]:
        pass

    def _build_pusher_args(self):
        return {'push_destination': self.build_push_destination(),
                'custom_config': self.get_config()}

    def _get_executor_spec(self):
        return executor_spec.ExecutorClassSpec(self.get_executor())



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Tests for different split steps."""

import random

import pytest
import tensorflow as tf

from zenml.steps.split import CategoricalDomainSplit
from zenml.steps.split import CategoricalRatioSplit
from zenml.steps.split import NoSplit
from zenml.steps.split import RandomSplit


@pytest.fixture
def create_random_dummy_data():
    def create_data():
        cat_col = "my_cat_col"
        possible_values = ["value{}".format(i + 1) for i in range(3)]
        dummy_data = []

        for i in range(10):
            value = random.choice(possible_values).encode()
            feature = {cat_col: tf.train.Feature(
                bytes_list=tf.train.BytesList(value=[value]))}

            dummy_data.append(tf.train.Example(
                features=tf.train.Features(feature=feature)))

        return dummy_data

    return create_data


@pytest.fixture
def create_structured_dummy_data():
    def create_data(counts):

        cat_col = "my_cat_col"
        possible_values = ["value{}".format(i + 1) for i in range(len(counts))]
        dummy_data = []

        for i, nums in enumerate(counts):
            value = possible_values[i].encode()
            feature = {cat_col: tf.train.Feature(
                bytes_list=tf.train.BytesList(value=[value]))}
            for n in range(nums):
                dummy_data.append(tf.train.Example(
                    features=tf.train.Features(feature=feature)))

        return dummy_data

    return create_data


def test_no_split(create_random_dummy_data):
    nosplit = NoSplit()

    # test defaults
    assert not nosplit.schema
    assert not nosplit.statistics

    dummy_data = create_random_dummy_data()

    split_folds = [nosplit.partition_fn(ex, nosplit.get_num_splits())
                   for ex in dummy_data]

    # assert nosplit returns only one index
    assert all(fold == 0 for fold in split_folds)


def test_random_split(create_random_dummy_data):
    one_fold = {"train": 1.0}

    # only one argument present in split map
    with pytest.raises(AssertionError):
        _ = RandomSplit(split_map=one_fold)

    bogus_entries = {"train": 0.5,
                     "eval": "testest"}

    # not all entries in split map are floats
    with pytest.raises(AssertionError):
        _ = RandomSplit(split_map=bogus_entries)

    split_map = {"train": 1.0,
                 "eval": 0.0}

    random_split = RandomSplit(split_map=split_map)

    # test defaults
    assert not random_split.schema
    assert not random_split.statistics

    dummy_data = create_random_dummy_data()

    split_folds = [random_split.partition_fn(ex, random_split.get_num_splits())
                   for ex in dummy_data]

    # artificial no split result tests, everything else is random
    assert all(fold == 0 for fold in split_folds)


def test_categorical_domain_split(create_structured_dummy_data):
    cat_col = "my_cat_col"

    one_fold = {"train": []}

    # only one argument present in split map
    with pytest.raises(AssertionError):
        _ = CategoricalDomainSplit(categorical_column=cat_col,
                                   split_map=one_fold)

    # real logic begins here
    split_map = {"train": ["value1"],
                 "eval": ["value2"],
                 "test": ["value3"]}

    domain_split = CategoricalDomainSplit(categorical_column=cat_col,
                                          split_map=split_map)

    # test defaults
    assert not domain_split.schema
    assert not domain_split.statistics

    # each categorical value in the split map above gets one example
    counts = [1, 1, 1]

    dummy_data = create_structured_dummy_data(counts)

    split_folds = [domain_split.partition_fn(ex, domain_split.get_num_splits())
                   for ex in dummy_data]

    # fold indices, zero-based, should correspond to their dict counterparts
    assert split_folds == [0, 1, 2]

    # each categorical value in the split map above gets one example,
    # plus one out-of-split-map (unseen example)
    counts = [1, 1, 1, 1]

    dummy_data = create_structured_dummy_data(counts)

    split_folds = [domain_split.partition_fn(ex, domain_split.get_num_splits())
                   for ex in dummy_data]

    assert domain_split.get_num_splits() == 4

    # default behavior is skipping (index n-1 = 3)
    assert split_folds == [0, 1, 2, 3]

    # test whether eval assignment works
    domain_split.unknown_category_policy = "eval"

    # each categorical value in the split map above gets one example,
    # plus one out-of-split-map (unseen example)
    counts = [1, 1, 1, 1]

    dummy_data = create_structured_dummy_data(counts)

    split_folds = [domain_split.partition_fn(ex, domain_split.get_num_splits())
                   for ex in dummy_data]

    assert domain_split.get_num_splits() == 3

    # default behavior is eval (index 1)
    assert split_folds == [0, 1, 2, 1]


def test_categorical_ratio_split(create_structured_dummy_data):
    cat_col = "my_cat_col"

    categories = []

    one_fold = {"train": 1.0}

    # only one fold present in split map
    with pytest.raises(AssertionError):
        _ = CategoricalRatioSplit(categorical_column=cat_col,
                                  categories=categories,
                                  split_ratio=one_fold)

    # real logic begins here
    categories = ["value{}".format(i + 1) for i in range(3)]

    split_ratio = {"train": 0.33,
                   "eval": 0.33,
                   "test": 0.34}

    ratio_split = CategoricalRatioSplit(categorical_column=cat_col,
                                        categories=categories,
                                        split_ratio=split_ratio)

    # test defaults
    assert not ratio_split.schema
    assert not ratio_split.statistics

    # each categorical value in the split map above gets one example
    counts = [1, 1, 1]

    dummy_data = create_structured_dummy_data(counts)

    split_folds = [ratio_split.partition_fn(ex, ratio_split.get_num_splits())
                   for ex in dummy_data]

    # fold indices, zero-based, should correspond to their dict counterparts
    assert split_folds == [0, 1, 2]

    # each categorical value in the split map above gets one example,
    # plus one out-of-split-map (unseen example)
    counts = [1, 1, 1, 1]

    dummy_data = create_structured_dummy_data(counts)

    split_folds = [ratio_split.partition_fn(ex, ratio_split.get_num_splits())
                   for ex in dummy_data]

    assert ratio_split.get_num_splits() == 4

    # default behavior is assigning everything into eval (index 1)
    assert split_folds == [0, 1, 2, 3]

    # test whether eval assignment works
    ratio_split.unknown_category_policy = "eval"

    # each categorical value in the split map above gets one example,
    # plus one out-of-split-map (unseen example)
    counts = [1, 1, 1, 1]

    dummy_data = create_structured_dummy_data(counts)

    split_folds = [ratio_split.partition_fn(ex, ratio_split.get_num_splits())
                   for ex in dummy_data]

    assert ratio_split.get_num_splits() == 3

    # default behavior is eval (index 1)
    assert split_folds == [0, 1, 2, 1]


def test_categorical_split_ordering(create_structured_dummy_data):
    cat_col = "my_cat_col"

    # real logic begins here
    split_map = {"train": ["value1"],
                 "eval": ["value2"]}

    domain_split = CategoricalDomainSplit(categorical_column=cat_col,
                                          split_map=split_map)

    # each categorical value in the split map above gets one example
    counts = [3, 1]

    dummy_data = create_structured_dummy_data(counts)

    split_folds = [domain_split.partition_fn(ex, domain_split.get_num_splits())
                   for ex in dummy_data]

    # expected: first 3 are 0 (value1, going into train),
    # last is 1 (value2, going into eval)
    assert split_folds == [0, 0, 0, 1]
    assert domain_split.get_split_names() == ["train", "eval", "skip"]

    ################################################
    # NOW: Order reversed, eval comes before train #
    ################################################

    # same split map, fold orders reversed
    split_map = {"eval": ["value2"],
                 "train": ["value1"]}

    domain_split = CategoricalDomainSplit(categorical_column=cat_col,
                                          split_map=split_map)

    # value_1 gets 3 examples, value_2 gets 1 example
    counts = [3, 1]

    dummy_data = create_structured_dummy_data(counts)

    split_folds = [domain_split.partition_fn(ex, domain_split.get_num_splits())
                   for ex in dummy_data]

    # expected: first 3 are 1 (value1, going into train), last is 0 (eval)
    assert split_folds == [1, 1, 1, 0]
    assert domain_split.get_split_names() == ["eval", "train", "skip"]



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import enum
import pprint

import yaml
from dateutil import tz


class PrintStyles(enum.Enum):
    NATIVE = 0
    YAML = 1
    PPRINT = 2


def to_pretty_string(d, style=PrintStyles.YAML):
    """
    Args:
        d:
        style:
    """
    if style == PrintStyles.YAML:
        return yaml.dump(d, default_flow_style=False)
    elif style == PrintStyles.PPRINT:
        return pprint.pformat(d)
    else:
        return str(d)


def format_date(dt, format='%Y-%m-%d %H:%M:%S'):
    """
    Formatting a datetime object nicely.

    Args:
        dt: datetime object
        format: format
    """
    if dt is None:
        return ''
    local_zone = tz.tzlocal()
    # make sure this is UTC
    dt = dt.replace(tzinfo=tz.tzutc())
    local_time = dt.astimezone(local_zone)
    return local_time.strftime(format)


def format_timedelta(td):
    """
    Format timedelta object nicely.

    Args:
        td: timedelta object
    """
    if td is None:
        return ''
    hours, remainder = divmod(td.total_seconds(), 3600)
    minutes, seconds = divmod(remainder, 60)
    return '{:02}:{:02}:{:02}'.format(int(hours), int(minutes), int(seconds))



import os

from zenml.datasources import CSVDatasource
from zenml.exceptions import AlreadyExistsException
from zenml.pipelines import TrainingPipeline
from zenml.repo import Repository, ArtifactStore
from zenml.steps.deployer import GCAIPDeployer
from zenml.steps.evaluator import TFMAEvaluator
from zenml.steps.preprocesser import StandardPreprocesser
from zenml.steps.split import RandomSplit
from zenml.steps.trainer import TFFeedForwardTrainer
from zenml.utils.naming_utils import transformed_label_name

GCP_PROJECT = os.getenv('GCP_PROJECT')
GCP_BUCKET = os.getenv('GCP_BUCKET')
MODEL_NAME = os.getenv('MODEL_NAME')

assert GCP_PROJECT
assert GCP_BUCKET
assert MODEL_NAME

# Deploy a tensorflow model on GCAIP. Note that no other trainer type
# works with this deployer except for the one shown here.

# Define the training pipeline
training_pipeline = TrainingPipeline()

# Add a datasource. This will automatically track and version it.
try:
    ds = CSVDatasource(name='Pima Indians Diabetes',
                       path='gs://zenml_quickstart/diabetes.csv')
except AlreadyExistsException:
    ds = Repository.get_instance().get_datasource_by_name(
        'Pima Indians Diabetes')
training_pipeline.add_datasource(ds)

# Add a split
training_pipeline.add_split(RandomSplit(
    split_map={'train': 0.7, 'eval': 0.2, 'test': 0.1}))

# Add a preprocessing unit
training_pipeline.add_preprocesser(
    StandardPreprocesser(
        features=['times_pregnant', 'pgc', 'dbp', 'tst', 'insulin', 'bmi',
                  'pedigree', 'age'],
        labels=['has_diabetes'],
        overwrite={'has_diabetes': {
            'transform': [{'method': 'no_transform', 'parameters': {}}]}}
    ))

training_pipeline.add_trainer(TFFeedForwardTrainer(
    loss='binary_crossentropy',
    last_activation='sigmoid',
    output_units=1,
    metrics=['accuracy'],
    epochs=20))

# Add an evaluator
training_pipeline.add_evaluator(
    TFMAEvaluator(slices=[['has_diabetes']],
                  metrics={transformed_label_name('has_diabetes'):
                               ['binary_crossentropy', 'binary_accuracy']}))

# Add the deployer
training_pipeline.add_deployment(
    GCAIPDeployer(
        project_id=GCP_PROJECT,
        model_name=MODEL_NAME,
    )
)

# Define the artifact store
artifact_store = ArtifactStore(
    os.path.join(GCP_BUCKET, 'gcp_gcaip_deployment/artifact_store'))

# Run the pipeline
training_pipeline.run(artifact_store=artifact_store)

# Another way to do is is to create a DeploymentPipeline.
# Uncomment to create the model via this pipeline
# from zenml.core.pipelines.deploy_pipeline import DeploymentPipeline
# model_uri = training_pipeline.get_model_uri()
# deploy_pipeline = DeploymentPipeline(model_uri=model_uri)
# deploy_pipeline.add_deployment(
#     GCAIPDeployer(
#         model_name=MODEL_NAME + '_v2',
#         project_id=GCP_PROJECT
#     )
# )
# deploy_pipeline.run(artifact_store=artifact_store)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


from typing import Dict, List, Text

import tensorflow as tf
import tensorflow_transform as tft

from zenml.steps.preprocesser import BasePreprocesserStep
from zenml.steps.preprocesser.standard_preprocesser.methods \
    .methods_nonseq_filling import \
    _impute
from zenml.utils import naming_utils


def decode_image(input_):
    densed = tf.sparse.to_dense(input_)
    image = tf.map_fn(lambda x: tf.io.decode_image(x[0], channels=3), densed,
                      dtype=tf.uint8)
    return image


def reshape_image(input_, shape):
    image = (tf.cast(input_, tf.float32) / 127.5) - 1
    image = tf.reshape(image, shape)
    return image


class ImagePreprocessor(BasePreprocesserStep):
    def __init__(self, shape: List = None, image_key: Text = 'image',
                 labels: List = None, **kwargs):
        if shape is None:
            shape = []
        if labels is None:
            labels = []
        self.shape = shape
        self.labels = labels
        self.image_key = image_key
        super().__init__(shape=shape, image_key=image_key, labels=labels,
                         **kwargs)

    def preprocessing_fn(self, inputs: Dict):

        outputs = {}

        for k, v in inputs.items():
            if k == self.image_key:
                result = decode_image(v)
                result = reshape_image(result, shape=self.shape)
                result = tf.cast(result, dtype=tf.float32)
                outputs[naming_utils.transformed_feature_name(k)] = result
            if k in self.labels:
                m = tft.max(v)
                result = _impute(v, m)
                # always retain raw data
                outputs[k] = result

                result = tf.cast(v, dtype=tf.float32)
                outputs[naming_utils.transformed_label_name(k)] = result


        return outputs



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import Text

import click
import git

from zenml.cli.cli import cli
from zenml.cli.utils import confirmation
from zenml.cli.utils import pass_repo
from zenml.repo import Repository
from zenml.utils.analytics_utils import track, INITIALIZE


@cli.command('init')
@click.option('--repo_path', type=click.Path(exists=True))
@click.option('--pipelines_dir', type=click.Path(exists=True))
@click.option('--analytics_opt_in', '-a', type=click.BOOL)
@track(event=INITIALIZE)
def init(repo_path: Text, pipelines_dir: Text = None,
         analytics_opt_in: bool = True):
    """Initialize ZenML on given path."""
    if repo_path is None:
        repo_path = os.getcwd()

    try:
        Repository.init_repo(
            repo_path,
            None,
            None,
            pipelines_dir,
            analytics_opt_in,
        )
        click.echo(f'ZenML repo initialized at {repo_path}')
    except git.InvalidGitRepositoryError:
        click.echo(f'{repo_path} is not a valid git repository! Please '
                   f'initialize ZenML within a git repository using '
                   f'`git init `')
    except AssertionError as e:
        click.echo(f'{e}')


@cli.command('clean')
@click.option('--yes', '-y', type=click.BOOL, default=False)
@pass_repo
def clean(repo: Repository, yes: bool = False):
    """Clean everything in repository."""
    if not yes:
        confirm = confirmation(
            "This will completely delete all pipelines, their associated "
            "artifacts and metadata ever created in this ZenML repository. "
            "Are you sure you want to proceed?")
    else:
        confirm = True

    click.echo("Not implemented for this version")
    # if confirm:
    #     repo.clean()



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


from typing import List, Text

import tensorflow as tf

from zenml.steps.trainer import TFBaseTrainerStep
from zenml.steps.trainer import utils
from zenml.utils import naming_utils


class FeedForwardTrainer(TFBaseTrainerStep):
    """
    Basic Feedforward Neural Network trainer. This step serves as an example
    of how to define your training logic to integrate well with TFX and
    Tensorflow Serving.
    """

    def __init__(self,
                 batch_size: int = 8,
                 lr: float = 0.001,
                 epochs: int = 1,
                 dropout_chance: int = 0.2,
                 loss: str = 'mse',
                 metrics: List[str] = None,
                 hidden_layers: List[int] = None,
                 hidden_activation: str = 'relu',
                 last_activation: str = 'sigmoid',
                 output_units: int = 1,
                 **kwargs):
        """
        Basic feedforward neural network constructor.

        Args:
            batch_size: Input data batch size.
            lr: Learning rate of the optimizer.
            epochs: Number of training epochs (whole passes through the data).
            dropout_chance: Dropout chance, i.e. probability of a neuron not
             propagating its weights into the prediction at a dropout layer.
            loss: Name of the loss function to use.
            metrics: List of metrics to log in training.
            hidden_layers: List of sizes to use in consecutive hidden layers.
             Length determines the number of hidden layers in the network.
            hidden_activation: Name of the activation function to use in the
             hidden layers.
            last_activation: Name of the final activation function creating
             the class probability distribution.
            output_units: Number of output units, corresponding to the number
             of classes.
            **kwargs: Additional keyword arguments.
        """

        self.batch_size = batch_size
        self.lr = lr
        self.epochs = epochs
        self.dropout_chance = dropout_chance
        self.loss = loss
        self.metrics = metrics or []
        self.hidden_layers = hidden_layers or [10]
        self.hidden_activation = hidden_activation
        self.last_activation = last_activation
        self.output_units = output_units

        super(FeedForwardTrainer, self).__init__(
            batch_size=batch_size,
            lr=lr,
            epochs=epochs,
            dropout_chance=dropout_chance,
            loss=loss,
            metrics=metrics,
            hidden_layers=hidden_layers,
            hidden_activation=hidden_activation,
            last_activation=last_activation,
            output_units=output_units,
            **kwargs
        )

    def test_fn(self, model, dataset):
        batch_list = []
        for x, y in dataset:
            # start with an empty batch
            batch = {}

            transformed_f = {f: x[f] for f in x if
                             naming_utils.check_if_transformed_feature(f)}
            raw_f = {f: x[f] for f in x if
                     not naming_utils.check_if_transformed_feature(f)}

            # add the raw features with the transformed features and labels
            batch.update(transformed_f)
            batch.update(y)
            batch.update(raw_f)

            # finally, add the output of the
            p = model.predict(x)

            if isinstance(p, tf.Tensor):
                batch.update({'output': p})
            elif isinstance(p, dict):
                batch.update({naming_utils.output_name(k): p[k] for k in p})
            elif isinstance(p, list):
                batch.update(
                    {'output_{}'.format(i): v for i, v in enumerate(p)})
            else:
                raise TypeError('Unknown output format!')

            batch_list.append(batch)

        combined_batch = utils.combine_batch_results(batch_list)

        return combined_batch

    def run_fn(self):
        train_split_patterns = [self.input_patterns[split] for split in
                                self.split_mapping[utils.TRAIN_SPLITS]]
        train_dataset = self.input_fn(train_split_patterns)

        eval_split_patterns = [self.input_patterns[split] for split in
                               self.split_mapping[utils.EVAL_SPLITS]]
        eval_dataset = self.input_fn(eval_split_patterns)

        model = self.model_fn(train_dataset=train_dataset,
                              eval_dataset=eval_dataset)

        for split in self.split_mapping[utils.TEST_SPLITS]:
            assert split in self.input_patterns, \
                f'There are currently no inputs for the split "{split}" ' \
                f'which is currently used in the {utils.TEST_SPLITS} of the ' \
                f'split mapping.'
            pattern = self.input_patterns[split]
            test_dataset = self.input_fn([pattern])
            test_results = self.test_fn(model, test_dataset)
            utils.save_test_results(test_results, self.output_patterns[split])

        signatures = self.get_signatures(model)

        model.save(self.serving_model_dir,
                   save_format='tf',
                   signatures=signatures)

    def input_fn(self,
                 file_pattern: List[Text]):
        """
        Feedforward input_fn for loading data from TFRecords saved to a
        location on disk.

        Args:
            file_pattern: File pattern matching saved TFRecords on disk.

        Returns:
            dataset: tf.data.Dataset created out of the input files.
        """
        dataset = tf.data.experimental.make_batched_features_dataset(
            file_pattern=file_pattern,
            batch_size=self.batch_size,
            features=self.schema,
            reader=self._gzip_reader_fn,
            num_epochs=1,
            drop_final_batch=True)

        def split_columns(x):
            inputs = {}
            labels = {}
            for e in x:
                if not naming_utils.check_if_transformed_label(e):
                    inputs[e] = x[e]
                else:
                    labels[e] = x[e]
            return inputs, labels

        dataset = dataset.map(split_columns)

        return dataset

    def get_signatures(self, model):
        return {
            'serving_default':
                self._get_serve_tf_examples_fn(
                    model,
                    self.tf_transform_output
                ).get_concrete_function(tf.TensorSpec(shape=[None],
                                                      dtype=tf.string,
                                                      name='examples')),
            'zen_eval':
                self._get_zen_eval_tf_examples_fn(
                    model,
                    self.tf_transform_output
                ).get_concrete_function(tf.TensorSpec(shape=[None],
                                                      dtype=tf.string,
                                                      name='examples'))}

    @staticmethod
    def _get_serve_tf_examples_fn(model, tf_transform_output):
        """
        Returns a function that parses a serialized tf.Example for prediction.

        Args:
            model: Trained model, output of the model_fn.
            tf_transform_output: Output of the preceding Transform /
             Preprocessing component.
        """

        model.tft_layer = tf_transform_output.transform_features_layer()

        @tf.function
        def serve_tf_examples_fn(serialized_tf_examples):
            """Returns the output to be used in the serving signature."""
            raw_feature_spec = tf_transform_output.raw_feature_spec()
            parsed_features = tf.io.parse_example(serialized_tf_examples,
                                                  raw_feature_spec)

            transformed_features = model.tft_layer(parsed_features)

            return model(transformed_features)

        return serve_tf_examples_fn

    @staticmethod
    def _get_zen_eval_tf_examples_fn(model, tf_transform_output):
        """
        Returns a function that parses a serialized tf.Example for
        evaluation.

        Args:
            model: Trained model, output of the model_fn.
            tf_transform_output: Output of the preceding Transform /
             Preprocessing component.
        """

        model.tft_layer = tf_transform_output.transform_features_layer()

        @tf.function
        def zen_eval_tf_examples_fn(serialized_tf_examples):
            """Returns the output to be used in the ce_eval signature."""
            xf_feature_spec = tf_transform_output.transformed_feature_spec()

            transformed_features = tf.io.parse_example(serialized_tf_examples,
                                                       xf_feature_spec)

            for f in tf_transform_output.transformed_feature_spec():
                if not naming_utils.transformed_feature_name(f):
                    transformed_features.pop(f)

            outputs = model(transformed_features)

            if isinstance(outputs, dict):
                return outputs
            elif isinstance(outputs, list):
                return {o.name.split("/")[1]: o for o in outputs}
            elif isinstance(outputs, tf.Tensor):
                return {outputs.name.split("/")[1]: outputs}

        return zen_eval_tf_examples_fn

    @staticmethod
    def _gzip_reader_fn(filenames):
        """
        Small utility returning a record reader that can read gzipped files.

        Args:
            filenames: Names of the compressed TFRecord data files.
        """
        return tf.data.TFRecordDataset(filenames, compression_type='GZIP')

    def model_fn(self,
                 train_dataset: tf.data.Dataset,
                 eval_dataset: tf.data.Dataset):
        """
        Function defining the training flow of the feedforward neural network
        model.

        Args:
            train_dataset: tf.data.Dataset containing the training data.
            eval_dataset: tf.data.Dataset containing the evaluation data.

        Returns:
            model: A trained feedforward neural network model.
        """

        features = list(train_dataset.element_spec[0].keys())
        labels = list(train_dataset.element_spec[1].keys())

        input_layers = [tf.keras.layers.Input(shape=(1,), name=k)
                        for k in features if
                        naming_utils.check_if_transformed_feature(k)]

        d = tf.keras.layers.Concatenate()(input_layers)

        for size in self.hidden_layers:
            d = tf.keras.layers.Dense(size,
                                      activation=self.hidden_activation)(d)
            d = tf.keras.layers.Dropout(self.dropout_chance)(d)

        output_layers = {l: tf.keras.layers.Dense(
            self.output_units,
            activation=self.last_activation,
            name=l)(d) for l in labels}

        model = tf.keras.Model(inputs=input_layers, outputs=output_layers)

        model.compile(loss=self.loss,
                      optimizer=tf.keras.optimizers.Adam(lr=self.lr),
                      metrics=self.metrics)

        model.summary()

        tensorboard_callback = tf.keras.callbacks.TensorBoard(
            log_dir=self.log_dir)

        model.fit(train_dataset,
                  epochs=self.epochs,
                  validation_data=eval_dataset,
                  callbacks=[tensorboard_callback])

        return model



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""CSv Datasource definition"""

import os
from typing import Callable, Text, Dict

import apache_beam as beam
from tfx_bsl.coders import csv_decoder

from zenml.datasources.base_datasource import BaseDatasource
from zenml.logger import get_logger
from zenml.utils import path_utils
from zenml.utils.beam_utils import WriteToTFRecord

logger = get_logger(__name__)


class CSVDatasource(BaseDatasource):
    """ZenML CSV datasource definition.

    Use this for CSV training pipelines.
    """

    def __init__(
            self,
            name: Text,
            path: Text,
            schema: Dict = None,
            **kwargs):
        """
        Create a CSV datasource. Creating this datasource creates a Beam
        pipeline that converts the CSV file into TFRecords for pipelines to
        use.

        The path can be a local path or a Google Cloud Storage bucket
        path for now (S3, Azure coming soon). The path defines the datasource,
        meaning a change in it (including file name) should be dealt with by
        creating another datasource.

        Args:
            name (str): name of datasource.
            path (str): path to csv file.
            schema (str): optional schema for data to conform to.
        """
        self.path = path
        self.schema = schema
        super().__init__(name, path=path, schema=schema, **kwargs)

    def process(self, output_path: Text, make_beam_pipeline: Callable = None):
        wildcard_qualifier = "*"
        file_pattern = os.path.join(self.path, wildcard_qualifier)

        if path_utils.is_dir(self.path):
            csv_files = path_utils.list_dir(self.path)
            if not csv_files:
                raise RuntimeError(
                    'Split pattern {} does not match any files.'.format(
                        file_pattern))
        else:
            if path_utils.file_exists(self.path):
                csv_files = [self.path]
            else:
                raise RuntimeError(f'{self.path} does not exist.')

        # weed out bad file exts with this logic
        allowed_file_exts = [".csv", ".txt"]  # ".dat"
        csv_files = [uri for uri in csv_files if os.path.splitext(uri)[1]
                     in allowed_file_exts]

        logger.info(f'Matched {len(csv_files)}: {csv_files}')

        # Always use header from file
        logger.info(f'Using header from file: {csv_files[0]}.')
        column_names = path_utils.load_csv_header(csv_files[0])
        logger.info(f'Header: {column_names}.')

        with make_beam_pipeline() as p:
            p | 'ReadFromText' >> beam.io.ReadFromText(
                file_pattern=self.path,
                skip_header_lines=1) \
            | 'ParseCSVLine' >> beam.ParDo(csv_decoder.ParseCSVLine(
                delimiter=',')) \
            | 'ExtractParsedCSVLines' >> beam.Map(
                lambda x: dict(zip(column_names, x[0]))) \
            | WriteToTFRecord(self.schema, output_path)



import tensorflow as tf


# The following functions can be used to convert a value to a type compatible
# with tf.train.Example.

def _bytes_feature(value):
    """Returns a bytes_list from a string / byte."""
    if isinstance(value, type(tf.constant(0))):
        value = value.numpy()  # BytesList won't unpack a string from an
        # EagerTensor.
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def _float_feature(value):
    """Returns a float_list from a float / double."""
    return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))


def _int64_feature(value):
    """Returns an int64_list from a bool / enum / int / uint."""
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def prepare_tf_example(raw_dict):
    feature = {}
    for k, v in raw_dict.items():
        if type(v) == int:
            feature[k] = _int64_feature(v)
        elif type(v) == float:
            feature[k] = _float_feature(v)
        else:
            feature[k] = _bytes_feature(str(v).encode())

    # return the TFExample
    return tf.train.Example(features=tf.train.Features(feature=feature))


class TensorFlowPredictor:
    def __init__(self, tensorflow_client, config):
        print("Got config to be %s" % str(config))
        self.client = tensorflow_client
        self.model_path = config['model_artifact']

    def predict(self, payload, query_params):
        print("Got payload to be %s" % str(payload))
        print("Got query_params to be %s" % str(query_params))

        model_name = None
        if 'model' in query_params:
            model_name = query_params['model']

        # Convert to TFExample because we used the TensorflowTrainer
        tf_example = prepare_tf_example(payload)
        model_input = {
            "examples": [tf_example.SerializeToString()]
        }
        # do a forward pass with the model
        return self.client.predict(model_input, model_name)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""JSON Datasource definition"""

from typing import Text, Callable, Dict
import json
import apache_beam as beam

from zenml.datasources import BaseDatasource
from zenml.utils.beam_utils import WriteToTFRecord
from zenml.utils import path_utils


class JSONDatasource(BaseDatasource):
    """ZenML JSON datasource definition."""

    def __init__(
            self,
            name: Text,
            path: Text,
            schema: Dict = None,
            **kwargs):
        """
        Initialize JSON datasource.

        Args:
            name: Name of datasource.
            path: path to json file, has to be structured as a list of dicts.
            schema (str): optional schema for data to conform to.
        """
        self.path = path
        self.schema = schema
        super().__init__(name, path=path, schema=schema, **kwargs)

    def process(self, output_path: Text, make_beam_pipeline: Callable = None):
        contents = path_utils.read_file_contents(self.path)
        json_obj = json.loads(contents)
        with make_beam_pipeline() as p:
            (p
             | beam.Create(json_obj)
             | WriteToTFRecord(self.schema, output_path)
             )



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from datetime import datetime
from typing import Dict, Text

import apache_beam as beam
import pandas as pd
import pytz
from apache_beam.transforms.window import Sessions
from apache_beam.transforms.window import TimestampedValue
from apache_beam.utils.timestamp import Timestamp
from tensorflow_transform.tf_metadata import schema_utils

from zenml.steps.sequencer import BaseSequencerStep
from zenml.steps.sequencer.standard_sequencer import utils
from zenml.steps.sequencer.standard_sequencer.methods.standard_methods \
    import ResamplingMethods, FillingMethods
from zenml.utils.preprocessing_utils import parse_methods, DEFAULT_DICT


class StandardSequencer(BaseSequencerStep):

    def __init__(self,
                 timestamp_column: Text,
                 category_column: Text = None,
                 overwrite: Dict[Text, Text] = None,
                 resampling_freq: Text = '1D',
                 gap_threshold: int = 60000,
                 sequence_length: int = 4,
                 sequence_shift: int = 1,
                 **kwargs):
        """
        Initializing the StandardSequencer step, which is responsible for
        extracting sequences from any timeseries dataset.

        The main logic behind this step can be summed up in a few steps
        as follows:

        1. First, we define how to add the corresponding timestamp to
        each datapoint by using the function `get_timestamp_do_fn`. In this
        implementation, the timestamp is expected to be a unix timestamp.

        2. Similarly, we define how to add a categorical key to each
        datapoint if a categorical column is provided.

        3. Following that, we use the timestamp, the categorical key and
        the gap threshold to split the data into so-called 'sessions'

        4. Once the data is split into sessions, we resample the sessions
        based on the `resampling_freq` to create equidistant timestamps,
        fill the missing values and extract the finalized sequences based on
        the `sequence_length` and `sequence_shift`

        :param timestamp_column: string, the name of the column for the
        timestamp resides
        :param category_column: string, the name of the column of a possible
        categorical feature
        :param overwrite: dict, used to overwrite any of the default resampling
        and filling behaviour
        :param resampling_freq: string, the resampling frequency as an
        Offset Alias
        :param gap_threshold: int, the minimum gap between two sessions in
        seconds
        :param sequence_length: int, the desired length of a sequence in terms
        of datapoints
        :param sequence_shift: int, the number steps to shift before extracting
        the next sequence
        :param kwargs: additional params
        """

        self.timestamp_column = timestamp_column
        self.category_column = category_column
        self.overwrite = overwrite
        self.sequence_length = sequence_length
        self.sequence_shift = sequence_shift
        self.resampling_freq = resampling_freq
        self.gap_threshold = gap_threshold

        self.resampling_dict = parse_methods(overwrite or {},
                                             'resampling',
                                             ResamplingMethods)

        self.resampling_default_dict = parse_methods(DEFAULT_DICT,
                                                     'resampling',
                                                     ResamplingMethods)

        self.filling_dict = parse_methods(overwrite or {},
                                          'filling',
                                          FillingMethods)

        self.filling_default_dict = parse_methods(DEFAULT_DICT,
                                                  'filling',
                                                  FillingMethods)

        super(StandardSequencer, self).__init__(
            timestamp_column=timestamp_column,
            category_column=category_column,
            overwrite=overwrite,
            resampling_freq=resampling_freq,
            gap_threshold=gap_threshold,
            sequence_length=sequence_length,
            sequence_shift=sequence_shift,
            **kwargs)

    def get_category_do_fn(self):
        """
        Creates a class which inherits from beam.DoFn to add a categorical key
        to each datapoint

        :return: an instance of the beam.DoFn
        """
        default_category = '__default__'
        category_column = self.category_column

        class AddKey(beam.DoFn):
            def process(self, element):
                """
                Add the category as a key to each element
                """
                if category_column:
                    element = element.set_index(
                        element[category_column].apply(
                            lambda x: x[0].decode('utf-8')),
                        drop=False)
                else:
                    element[default_category] = default_category
                    element = element.set_index(default_category)
                return element.iterrows()

        return AddKey()

    def get_timestamp_do_fn(self):
        """
        Creates a class which inherits from beam.DoFn to add the timestamp
        to each datapoint

        :return: an instance of the beam.DoFn
        """

        timestamp_column = self.timestamp_column

        class AddTimestamp(beam.DoFn):
            def process(self, element):
                """
                Parse the timestamp and add it to the datapoints
                """
                t = element[1][timestamp_column][0]
                timestamp = datetime.fromtimestamp(t, pytz.utc)
                unix_timestamp = Timestamp.from_utc_datetime(timestamp)
                yield TimestampedValue(element, unix_timestamp)

        return AddTimestamp()

    def get_window(self):
        """
        Returns a selected beam windowing strategy

        :return: the selected windowing strategy
        """
        return Sessions(self.gap_threshold)

    def get_combine_fn(self):
        """
        Creates a class which inherits from beam.CombineFn which processes
        sessions and extracts sequences from it

        :return: an instance of the beam.CombineFn
        """
        epoch_suffix = '_epoch'

        timestamp_column = self.timestamp_column
        category_column = self.category_column
        sequence_length = self.sequence_length
        sequence_shift = self.sequence_shift
        resampling_freq = self.resampling_freq
        resampling_dict = self.resampling_dict
        resampling_default_dict = self.resampling_default_dict
        filling_dict = self.filling_dict
        filling_default_dict = self.filling_default_dict

        schema = self.schema

        class SequenceCombine(beam.CombineFn):
            def create_accumulator(self, *args, **kwargs):
                return pd.DataFrame()

            def add_input(self, mutable_accumulator, element, *args, **kwargs):
                return mutable_accumulator.append(element)

            def merge_accumulators(self, accumulators, *args, **kwargs):
                return pd.concat(accumulators)

            def extract_output(self, accumulator, *args, **kwargs):
                # Get the required args
                spec = schema_utils.schema_as_feature_spec(schema).feature_spec
                schema_dict = utils.infer_schema(spec)

                # Extract the values from the arrays within the cells
                session = accumulator.applymap(utils.array_to_value)

                # Extract the category if there is one
                cat = None
                if category_column:
                    cat = session[category_column].unique()[0].decode('utf-8')

                # Set the timestamp as the index
                session[timestamp_column] = session[timestamp_column].apply(
                    lambda x: pd.to_datetime(datetime.fromtimestamp(x,
                                                                    pytz.utc)))
                session = session.set_index(timestamp_column)

                # RESAMPLING
                r_config = utils.fill_defaults(resampling_dict,
                                               resampling_default_dict,
                                               schema_dict)

                resample_functions = utils.get_function_dict(
                    r_config, ResamplingMethods)

                for key, function_list in resample_functions.items():
                    assert len(function_list) == 1, \
                        'Please only define a single function for resampling'
                    resample_functions[key] = function_list[0]

                feature_list = list(resample_functions.keys())
                for feature in feature_list:
                    if feature not in session:
                        session[feature] = None

                resampler = session[feature_list].resample(resampling_freq)
                session = resampler.agg(resample_functions)

                # TODO: Investigate the outcome of disabling this
                # session = session.dropna(subset=config_dict[cts.FEATURES],
                #                          how='all',
                #                          axis=0)

                # Resolving the dtype conflicts after resampling
                output_dt_map = utils.get_resample_output_dtype(
                    resampling_dict,
                    schema_dict)
                session = session.astype(output_dt_map)

                # FILLING
                f_config = utils.fill_defaults(filling_dict,
                                               filling_default_dict,
                                               schema_dict)

                filling_functions = utils.get_function_dict(f_config,
                                                            FillingMethods)

                for key, function_list in filling_functions.items():
                    assert len(function_list) == 1, \
                        'Please only define a single function for filling'
                    filling_functions[key] = function_list[0]
                session = session.agg(filling_functions)

                # Required Keys
                if category_column:
                    session[category_column] = cat

                index_s = session.index.to_series()
                session[timestamp_column] = index_s.apply(
                    lambda x: x.strftime("%Y-%m-%d %H:%M:%S.%f %Z"))

                session[timestamp_column + epoch_suffix] = index_s.apply(
                    lambda x: utils.convert_datetime_to_secs(x))

                session.sort_index(inplace=True)

                return utils.extract_sequences(session,
                                               sequence_length,
                                               resampling_freq,
                                               sequence_shift)

        return SequenceCombine()



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Dict, Text, Any, List

from tfx import types
from tfx.dsl.components.base import base_executor
from tfx.types import artifact_utils

from zenml.components.data_gen.constants import DATA_SPLIT_NAME
from zenml.datasources.base_datasource import BaseDatasource
from zenml.standards.standard_keys import StepKeys
from zenml.utils import source_utils


class DataExecutor(base_executor.BaseExecutor):

    def Do(self,
           input_dict: Dict[Text, List[types.Artifact]],
           output_dict: Dict[Text, List[types.Artifact]],
           exec_properties: Dict[Text, Any]) -> None:
        """
        Args:
            input_dict:
            output_dict:
            exec_properties:
        """
        source = exec_properties[StepKeys.SOURCE]
        args = exec_properties[StepKeys.ARGS]
        name = exec_properties[StepKeys.NAME]

        c = source_utils.load_source_path_class(source)
        # TODO [LOW]: Passing _id makes it load
        datasource: BaseDatasource = c(name=name, _id='_id', **args)

        # Get output split path
        examples_artifact = artifact_utils.get_single_instance(
            output_dict[DATA_SPLIT_NAME])
        split_names = [DATA_SPLIT_NAME]
        examples_artifact.split_names = artifact_utils.encode_split_names(
            split_names)
        output_split_path = artifact_utils.get_split_uri(
            [examples_artifact], DATA_SPLIT_NAME)

        datasource.process(
            output_path=output_split_path,
            make_beam_pipeline=self._make_beam_pipeline)



# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ZenML bulk inferrer executor."""

from typing import Any, Dict, List, Text
from typing import Optional

import apache_beam as beam
from absl import logging
from tfx import types
from tfx.components.util import model_utils
from tfx.dsl.components.base import base_executor
from tfx.proto import bulk_inferrer_pb2
from tfx.types import artifact_utils
from tfx.utils import path_utils
from tfx_bsl.public.proto import model_spec_pb2
from tfx.components.bulk_inferrer.executor import _RunInference

from zenml.components.bulk_inferrer.utils import convert_to_dict
from zenml.components.bulk_inferrer.constants import MODEL, EXAMPLES, \
    MODEL_BLESSING, PREDICTIONS
from zenml.standards.standard_keys import StepKeys
from zenml.steps.inferrer import BaseInferrer
from zenml.utils import source_utils


class BulkInferrerExecutor(base_executor.BaseExecutor):
    """TFX bulk inferrer executor."""

    def Do(self, input_dict: Dict[Text, List[types.Artifact]],
           output_dict: Dict[Text, List[types.Artifact]],
           exec_properties: Dict[Text, Any]) -> None:
        """Runs batch inference on a given model with given input examples.
        Args:
          input_dict: Input dict from input key to a list of Artifacts.
            - examples: examples for inference.
            - model: exported model.
            - model_blessing: model blessing result, optional.
          output_dict: Output dict from output key to a list of Artifacts.
            - output: bulk inference results.
          exec_properties: A dict of execution properties.
            - model_spec: JSON string of bulk_inferrer_pb2.ModelSpec instance.
            - data_spec: JSON string of bulk_inferrer_pb2.DataSpec instance.
        Returns:
          None
        """
        self._log_startup(input_dict, output_dict, exec_properties)

        source = exec_properties[StepKeys.SOURCE]
        args = exec_properties[StepKeys.ARGS]
        c = source_utils.load_source_path_class(source)
        inferrer_step: BaseInferrer = c(**args)

        output_examples = artifact_utils.get_single_instance(
            output_dict[PREDICTIONS])

        if EXAMPLES not in input_dict:
            raise ValueError('\'examples\' is missing in input dict.')
        if MODEL not in input_dict:
            raise ValueError('Input models are not valid, model '
                             'need to be specified.')
        if MODEL_BLESSING in input_dict:
            model_blessing = artifact_utils.get_single_instance(
                input_dict['model_blessing'])
            if not model_utils.is_model_blessed(model_blessing):
                logging.info('Model on %s was not blessed', model_blessing.uri)
                return
        else:
            logging.info(
                'Model blessing is not provided, exported model will be '
                'used.')

        model = artifact_utils.get_single_instance(
            input_dict[MODEL])
        model_path = path_utils.serving_model_path(model.uri)
        logging.info('Use exported model from %s.', model_path)

        output_example_spec = bulk_inferrer_pb2.OutputExampleSpec(
            output_columns_spec=[bulk_inferrer_pb2.OutputColumnsSpec(
                predict_output=bulk_inferrer_pb2.PredictOutput(
                    output_columns=[bulk_inferrer_pb2.PredictOutputCol(
                        output_key=x,
                        output_column=f'{x}_label', ) for x in
                        inferrer_step.get_labels()]))])

        model_spec = bulk_inferrer_pb2.ModelSpec()
        saved_model_spec = model_spec_pb2.SavedModelSpec(
            model_path=model_path,
            tag=model_spec.tag,
            signature_name=model_spec.model_signature_name)
        inference_spec = model_spec_pb2.InferenceSpecType()
        inference_spec.saved_model_spec.CopyFrom(saved_model_spec)

        self._run_model_inference(
            output_example_spec,
            input_dict[EXAMPLES],
            output_examples,
            inference_spec,
            inferrer_step
        )

    def _run_model_inference(
            self,
            output_example_spec: bulk_inferrer_pb2.OutputExampleSpec,
            examples: List[types.Artifact],
            output_examples: Optional[types.Artifact],
            inference_endpoint: model_spec_pb2.InferenceSpecType,
            inferrer_step: BaseInferrer,
    ) -> None:
        """Runs model inference on given examples data.
        Args:
          output_example_spec: bulk_inferrer_pb2.OutputExampleSpec instance.
          examples: List of `standard_artifacts.Examples` artifacts.
          output_examples: Optional output `standard_artifacts.Examples`
          artifact.
          inference_endpoint: Model inference endpoint.
          inferrer_step: Inferrer step supplied in the infer pipeline config.
        """

        # TODO[LOW]: Rewrite this since there is only one split in the
        #  DataGen output
        example_uris = {}
        for example_artifact in examples:
            for split in artifact_utils.decode_split_names(
                    example_artifact.split_names):
                example_uris[split] = artifact_utils.get_split_uri(
                    [example_artifact], split)

        output_examples.split_names = artifact_utils.encode_split_names(
            sorted(example_uris.keys()))

        with self._make_beam_pipeline() as pipeline:
            for split, example_uri in example_uris.items():
                output_examples_split_uri = artifact_utils.get_split_uri(
                    [output_examples], split)
                inferrer_step.set_output_uri(output_examples_split_uri)
                logging.info('Path of output examples split `%s` is %s.',
                             split, output_examples_split_uri)
                _ = (
                    pipeline
                    | 'RunInference[{}]'.format(split) >>
                    _RunInference(example_uri, inference_endpoint)
                    | 'ConvertToDict[{}]'.format(split) >>
                    beam.Map(convert_to_dict, output_example_spec)
                    | 'WriteOutput[{}]'.format(split) >>
                    inferrer_step.write_inference_results())

            logging.info('Output examples written to %s.', output_examples.uri)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import tensorflow as tf


def one_hot_encode(custom_value):
    """
    Args:
        custom_value:
    """
    def apply(input_):
        return tf.one_hot(input_,
                          custom_value,
                          on_value=1, off_value=0,
                          axis=None,
                          dtype=None,
                          name=None)

    return apply


def decode_and_reshape_image(input_):
    """
    Args:
        input_: bytes feature, bytes representation of input image
    """
    image = tf.map_fn(
        lambda x: tf.io.decode_image(x[0], channels=3),
        input_,
        dtype=tf.uint8)
    # image = tf.image.decode_image(input_, channels=3)
    image = (tf.cast(image, tf.float32) / 127.5) - 1
    image = tf.reshape(image, [-1, 256, 256, 3])
    return image


def load_binary_image(image):
    """
    Args:
        image:
    """
    return tf.io.decode_image(image).numpy()



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""CLI for manipulating ZenML local and global config file."""

from typing import Text

import click

from zenml.cli.cli import cli
from zenml.cli.cli import pass_config
from zenml.cli.utils import parse_unknown_options
from zenml.repo import Repository


@cli.group()
def config():
    """Config group"""
    pass


# Analytics
@config.group()
def analytics():
    """Analytics for opt-in and opt-out"""
    pass


@analytics.command('opt-in',
                   context_settings=dict(ignore_unknown_options=True))
@pass_config
def opt_in(config):
    """Opt-in to analytics"""
    config.set_analytics_opt_in(True)
    click.echo('Opted in to analytics.')


@analytics.command('opt-out',
                   context_settings=dict(ignore_unknown_options=True))
@pass_config
def opt_out(config):
    """Opt-out of analytics"""
    config.set_analytics_opt_in(False)
    click.echo('Opted out of analytics.')


@config.command("list")
@click.pass_context
def list_config(ctx):
    """Print the current ZenML config to the command line"""
    ctx.invoke(get_metadata_store)
    ctx.invoke(get_artifact_store)
    ctx.invoke(get_pipelines_dir)


# Metadata Store
@config.group()
def metadata():
    """Utilities for metadata store"""
    pass


@metadata.command('set', context_settings=dict(ignore_unknown_options=True))
@click.argument('store_type', type=str)
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def set_metadata_store(store_type, args):
    """Set metadata store for local config."""

    try:
        parsed_args = parse_unknown_options(args)
    except AssertionError as e:
        click.echo(str(e))
        return

    # TODO: [LOW] Hard-coded
    config = {
        'type': store_type,
        'args': parsed_args
    }
    from zenml.metadata.metadata_wrapper import ZenMLMetadataStore

    store = ZenMLMetadataStore.from_config(config)
    repo: Repository = Repository.get_instance()
    repo.zenml_config.set_metadata_store(store)

    click.echo(f'Metadata store set to: {store.to_config()}')


@metadata.command('get')
def get_metadata_store():
    """Print metadata store from local config."""
    repo: Repository = Repository.get_instance()
    click.echo(f'Metadata store: '
               f'{repo.get_default_metadata_store().to_config()}')


# Artifact Store
@config.group()
def artifacts():
    """Utilities for artifact store"""
    pass


@artifacts.command('set')
@click.argument('path', type=click.Path())
def set_artifact_store(path: Text = None):
    """Change artifact store for local config."""
    repo: Repository = Repository.get_instance()
    repo.zenml_config.set_artifact_store(path)
    click.echo(f'Default artifact store updated to {path}')


@artifacts.command('get')
def get_artifact_store():
    """Print artifact store from local config."""
    repo: Repository = Repository.get_instance()
    click.echo(f'Default artifact store points to: '
               f'{repo.get_default_artifact_store().path}')


# Pipeline Directory
@config.group()
def pipelines():
    """Utilities for pipelines dir store"""
    pass


@pipelines.command('set')
@click.argument('path', type=click.Path())
def set_pipelines_dir(path: Text = None):
    """Change pipelines dir for local config."""
    repo: Repository = Repository.get_instance()
    repo.zenml_config.set_pipelines_dir(path)
    click.echo(f'Default pipelines dir updated to {path}')


@pipelines.command('get')
def get_pipelines_dir():
    """Print pipelines dir from local config."""
    repo: Repository = Repository.get_instance()
    click.echo(f'Default pipelines dir points to: '
               f'{repo.get_default_pipelines_dir()}')



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Functions to handle ZenML config"""

import os
from typing import Text, Optional, Dict, Type

from zenml.metadata import ZenMLMetadataStore
from zenml.repo import ArtifactStore
from zenml.repo.constants import ZENML_CONFIG_NAME, \
    ARTIFACT_STORE_DEFAULT_DIR, PIPELINES_DEFAULT_DIR_NAME, \
    ML_METADATA_SQLITE_DEFAULT_NAME, ZENML_DIR_NAME
from zenml.standards import standard_keys as keys
from zenml.utils import path_utils, yaml_utils
from zenml.exceptions import InitializationException

ARTIFACT_STORE_KEY = keys.GlobalKeys.ARTIFACT_STORE
METADATA_KEY = keys.GlobalKeys.METADATA_STORE
PIPELINES_DIR_KEY = 'pipelines_dir'


class ZenMLConfig:
    """ZenML config class to handle config operations.

    This is an internal class and should not be used by the user.
    """

    def __init__(self, repo_path: Text):
        """
        Construct class instance for ZenML Config.

        Args:
            repo_path (str): path to root of repository.
        """
        self.repo_path = repo_path
        if not ZenMLConfig.is_zenml_dir(self.repo_path):
            raise InitializationException(
                f'This is not a ZenML repository, as it does '
                f'not contain the {ZENML_CONFIG_NAME} '
                f'config file. Please initialize your repo '
                f'with `zenml init` with the ZenML CLI.')

        self.config_dir = os.path.join(repo_path, ZENML_DIR_NAME)
        self.config_path = os.path.join(self.config_dir, ZENML_CONFIG_NAME)

        self.raw_config = yaml_utils.read_yaml(self.config_path)

        # Load self vars in init to be clean
        self.metadata_store: Optional[ZenMLMetadataStore] = None
        self.artifact_store: Optional[ArtifactStore] = None
        self.pipelines_dir: Text = ''

        # Override these using load_config
        self.from_config(self.raw_config)

    @staticmethod
    def is_zenml_dir(path: Text):
        """
        Check if dir is a zenml dir or not.

        Args:
            path (str): path to the root.
        """
        config_path = os.path.join(path, ZENML_DIR_NAME, ZENML_CONFIG_NAME)
        if not path_utils.file_exists(config_path):
            return False
        return True

    @staticmethod
    def to_config(path: Text,
                  artifact_store_path: Text = None,
                  metadata_store: Optional[Type[ZenMLMetadataStore]] = None,
                  pipelines_dir: Text = None):
        """
        Creates a default .zenml config at path/zenml/.zenml_config.

        Args:
            path (str): path to a directory.
            metadata_store: metadata store definition.
            artifact_store_path (str): path where to store artifacts.
            pipelines_dir (str): path where to store pipeline configs.
        """
        config_dir_path = os.path.join(path, ZENML_DIR_NAME)
        config_path = os.path.join(config_dir_path, ZENML_CONFIG_NAME)

        if path_utils.file_exists(config_path):
            raise AssertionError(f'.zenml file already exists at '
                                 f'{config_path}. '
                                 f'Cannot replace. Please delete the '
                                 f'{config_dir_path} directory first.')

        # Create config dir
        path_utils.create_dir_if_not_exists(config_dir_path)

        if artifact_store_path is None:
            artifact_store_path = \
                os.path.join(config_dir_path, ARTIFACT_STORE_DEFAULT_DIR)
        else:
            # if provided, then resolve it absolutely
            artifact_store_path = path_utils.resolve_relative_path(
                artifact_store_path)

        # create artifact_store path
        path_utils.create_dir_if_not_exists(artifact_store_path)

        if metadata_store is None:
            uri = os.path.join(
                artifact_store_path, ML_METADATA_SQLITE_DEFAULT_NAME)
            from zenml.metadata import \
                SQLiteMetadataStore
            metadata_dict = SQLiteMetadataStore(uri).to_config()
        else:
            metadata_dict = metadata_store.to_config()

        if pipelines_dir is None:
            pipelines_dir = os.path.join(path, PIPELINES_DEFAULT_DIR_NAME)
        else:
            # if provided, still resolve
            pipelines_dir = path_utils.resolve_relative_path(pipelines_dir)

        path_utils.create_dir_if_not_exists(pipelines_dir)
        config_dict = {
            ARTIFACT_STORE_KEY: artifact_store_path,
            METADATA_KEY: metadata_dict,
            PIPELINES_DIR_KEY: pipelines_dir,
        }
        # Write initial config
        yaml_utils.write_yaml(config_path, config_dict)

    def from_config(self, config_dict: Dict):
        """
        Sets metadata and artifact_store variables

        Args:
            config_dict (dict): .zenml config object in dict format.
        """
        assert METADATA_KEY in config_dict
        assert ARTIFACT_STORE_KEY in config_dict
        assert PIPELINES_DIR_KEY in config_dict

        self.artifact_store = ArtifactStore(config_dict[ARTIFACT_STORE_KEY])
        self.metadata_store = ZenMLMetadataStore.from_config(
            config=config_dict[METADATA_KEY]
        )
        self.pipelines_dir = config_dict[PIPELINES_DIR_KEY]

    def get_metadata_store(self) -> ZenMLMetadataStore:
        """Get metadata store from config."""
        return self.metadata_store

    def get_artifact_store(self) -> ArtifactStore:
        """Get artifact store from config"""
        return self.artifact_store

    def get_pipelines_dir(self) -> Text:
        """Get absolute path of pipelines dir from config"""
        return self.pipelines_dir

    def set_artifact_store(self, artifact_store_path: Text):
        """
        Updates artifact store to point to path.

        Args:
            artifact_store_path: new path to artifact store
        """
        self.artifact_store = ArtifactStore(artifact_store_path)
        self.save()

    def set_metadata_store(self, metadata_store: ZenMLMetadataStore):
        """
        Updates artifact store to point to path.

        Args:
            metadata_store: metadata store
        """
        self.metadata_store = metadata_store
        self.save()

    def set_pipelines_dir(self, pipelines_dir: Text):
        """
        Updates artifact store to point to path.

        Args:
            pipelines_dir: new path to pipelines dir
        """
        path_utils.create_dir_if_not_exists(pipelines_dir)
        self.pipelines_dir = pipelines_dir
        self.save()

    def save(self):
        config_dict = {
            ARTIFACT_STORE_KEY: self.artifact_store.path,
            METADATA_KEY: self.metadata_store.to_config(),
            PIPELINES_DIR_KEY: self.pipelines_dir,
        }

        # Write initial config
        yaml_utils.write_yaml(self.config_path, config_dict)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML specific exception definitions"""


class InitializationException(Exception):
    """Raises exception when a function is run before zenml initialization."""

    def __init__(self,
                 message='ZenML config is none. Did you do `zenml init`?'):
        super().__init__(message)


class EmptyDatasourceException(Exception):
    """Raises exception when a datasource data is accessed without running
    an associated pipeline."""

    def __init__(self,
                 message='This datasource has not been used in any '
                         'pipelines, therefore the associated data has no '
                         'versions. Please use this datasouce in any ZenML '
                         'pipeline with `pipeline.add_datasource('
                         'datasource)`'):
        super().__init__(message)


class DoesNotExistException(Exception):
    """Raises exception when the `name` does not exist in the system but an
    action is being done that requires it to be present."""

    def __init__(self,
                 name='',
                 reason='',
                 message='{} does not exist! This might be due to: {}'):
        super().__init__(message.format(name, reason))


class AlreadyExistsException(Exception):
    """Raises exception when the `name` already exist in the system but an
    action is trying to create a resource with the same name."""

    def __init__(self,
                 name='',
                 resource_type=''):
        message = f'{resource_type} `{name}` already exists! Please use ' \
                  f'Repository.get_instance().get_{resource_type}_by_name' \
                  f'("{name}") to fetch it.'
        super().__init__(message)


class PipelineNotSucceededException(Exception):
    """Raises exception when trying to fetch artifacts from a not succeeded
    pipeline."""

    def __init__(self,
                 name='',
                 message='{} is not yet completed successfully.'):
        super().__init__(message.format(name))



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of a random split of the input data set."""

import bisect
from typing import Text, List, Dict, Any

import numpy as np

from zenml.steps.split import BaseSplitStep


def lint_split_map(split_map: Dict[Text, float]):
    """Small utility to lint the split_map"""
    if len(split_map) <= 1:
        raise AssertionError('Please specify more than 1 split name in the '
                             'split_map!')

    if not all(isinstance(v, (int, float)) for v in split_map.values()):
        raise AssertionError("Only int or float values are allowed when "
                             "specifying a random split!")


class RandomSplit(BaseSplitStep):
    """
    Random split. Use this to randomly split data based on a cumulative
    distribution function defined by a split_map dict.
    """

    def __init__(
            self,
            split_map: Dict[Text, float],
            statistics=None,
            schema=None,
    ):
        """
        Random split constructor.

        Randomly split the data based on a cumulative distribution function
        defined by split_map.

        Example usage:

        # Split data randomly, but evenly into train, eval and test

        >>> split = RandomSplit(
        ... split_map = {"train": 0.334,
        ...              "eval": 0.333,
        ...              "test": 0.333})

        Here, each data split gets assigned about one third of the probability
        mass. The split is carried out by sampling from the categorical
        distribution defined by the values p_i in the split map, i.e.

        P(index = i) = p_i, i = 1,...,n ;

        where n is the number of splits defined in the split map. Hence, the
        values in the split map must sum up to 1. For more information, see
        https://en.wikipedia.org/wiki/Categorical_distribution.


        Args:
            statistics: Parsed statistics from a preceding StatisticsGen.
            schema: Parsed schema from a preceding SchemaGen.
            split_map: A dict { split_name: percentage of data in split }.
        """

        lint_split_map(split_map)
        self.split_map = split_map

        super().__init__(statistics=statistics,
                         schema=schema,
                         split_map=split_map)

    def partition_fn(self,
                     element: Any,
                     num_partitions: int) -> int:
        """
        Function for a random split of the data; to be used in a beam.Partition.
        This function implements a simple random split algorithm by drawing
        integers from a categorical distribution defined by the values in
        split_map.

        Args:
            element: Data point, in format tf.train.Example.
            num_partitions: Number of splits, unused here.

        Returns:
            An integer n, where 0 ≤ n ≤ num_partitions - 1.
        """

        # calculates probability mass of each split
        probability_mass = np.cumsum(list(self.split_map.values()))
        max_value = probability_mass[-1]

        return bisect.bisect(probability_mass, np.random.uniform(0, max_value))

    def get_split_names(self) -> List[Text]:
        return list(self.split_map.keys())



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from training.trainer import UrduTrainer
from zenml.datasources import CSVDatasource
from zenml.exceptions import AlreadyExistsException
from zenml.pipelines import NLPPipeline
from zenml.repo import Repository
from zenml.steps.split import RandomSplit
from zenml.steps.tokenizer import HuggingFaceTokenizerStep

nlp_pipeline = NLPPipeline()

try:
    ds = CSVDatasource(name="My Urdu Text",
                       path="gs://zenml_quickstart/urdu_fake_news.csv")
except AlreadyExistsException:
    ds = Repository.get_instance().get_datasource_by_name(name="My Urdu Text")

nlp_pipeline.add_datasource(ds)

tokenizer_step = HuggingFaceTokenizerStep(
    text_feature="news",
    tokenizer="bert-wordpiece",
    vocab_size=3000)

nlp_pipeline.add_tokenizer(tokenizer_step=tokenizer_step)

nlp_pipeline.add_split(RandomSplit(
    split_map={'train': 0.7, 'eval': 0.2, 'test': 0.1}))

nlp_pipeline.add_trainer(
    UrduTrainer(
        model_name="distilbert-base-uncased",
        epochs=1,
        batch_size=32,
        learning_rate=5e-3)
)

nlp_pipeline.run()

# evaluate the model with the sentence "The earth is flat"
# which should (ideally) return FAKE_NEWS
nlp_pipeline.predict_sentence("دنیا سیدھی ہے")



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Dict, Text, List, Any

import tensorflow as tf

from zenml.steps.split.utils import get_categorical_value
from zenml.steps.tokenizer import BaseTokenizer
from zenml.steps.tokenizer.utils import tokenizer_map
from zenml.utils import path_utils


class HuggingFaceTokenizerStep(BaseTokenizer):
    """
    Base step for Tokenizer usage in NLP pipelines.

    This step can be used in Natural Language Processing (NLP) pipelines,
    where tokenization is a special form of preprocessing for text data in
    sequence form.

    This step was primarily designed to integrate the Huggingface ecosystem
    (transformers, tokenizers, datasets) into ZenML. Other tokenization
    libraries can work with this step, but are not guaranteed to work.
    """

    def __init__(self,
                 text_feature: Text,
                 tokenizer: Text,
                 tokenizer_params: Dict[Text, Any] = None,
                 skip_training: bool = False,
                 vocab_size: int = 30000,
                 min_frequency: int = 2,
                 sentence_length: int = 128,
                 special_tokens: List[Text] = None,
                 batch_size: int = 64,
                 **kwargs):
        """
        Tokenizer step constructor.

        Args:
            text_feature: String, name of the text feature in the data.
            tokenizer: String, name of the Huggingface tokenizer to use.
             Can be either one of "bert-wordpiece", "byte-level-bpe",
              "char-level-bpe", "sentencepiece-bpe", "sentencepiece-unigram".
            tokenizer_params: Keyword arguments to be used in construction of
             the tokenizer given by the tokenizer argument.
            skip_training: Boolean indicating whether or not to skip training
             the tokenizer in the pipeline.
            vocab_size: Size of the target vocabulary.
            min_frequency: Minimal frequency of an element to be included into
             the vocabulary.
            sentence_length: Sentence length. Varies depending on the target
             model, usually 128 for BERT-based models. Longer sentences will
             be padded, shorter sentences will be truncated to match length.
            special_tokens: List of special tokens to include.
            batch_size: Batch size to use when training the tokenizer from an
             iterator.
            **kwargs: Additional keyword arguments.
        """

        super(HuggingFaceTokenizerStep, self).__init__(
            text_feature=text_feature,
            skip_training=skip_training,
            vocab_size=vocab_size,
            min_frequency=min_frequency,
            sentence_length=sentence_length,
            tokenizer=tokenizer,
            tokenizer_params=tokenizer_params,
            special_tokens=special_tokens,
            batch_size=batch_size,
            **kwargs)

        # training arguments for tokenizer
        self.vocab_size = vocab_size
        self.min_frequency = min_frequency
        self.special_tokens = special_tokens or [
            "[PAD]",
            "[UNK]",
            "[CLS]",
            "[SEP]",
            "[MASK]",
        ]

        # tokenizer config sub-block
        self.tokenizer_name = tokenizer
        self.tokenizer_params = tokenizer_params or {}
        self.sentence_length = sentence_length
        self.batch_size = batch_size
        self.tokenizer = tokenizer_map.get(
            self.tokenizer_name)(**self.tokenizer_params)

        # truncation and padding constrain length to sentence_length
        self.tokenizer.enable_padding(length=self.sentence_length)
        self.tokenizer.enable_truncation(max_length=self.sentence_length)

    def train(self, files: List[Text]):
        """
        Method for training a tokenizer on a data iterator. The iterator is
        constructed inside the function in a closure from a TFRecord dataset.

        Args:
            files: List of TFRecord files in GZIP format to ingest.
        """
        ds = tf.data.TFRecordDataset(files, compression_type="GZIP")

        ds_numpy = ds.batch(self.batch_size).as_numpy_iterator()

        def build_scoped_iterator(text_feature):
            for batch in ds_numpy:
                yield [get_categorical_value(
                    tf.train.Example.FromString(ex),
                    text_feature) for ex in batch]

        iterator = build_scoped_iterator(self.text_feature)

        # showing progress does not work in logging mode apparently
        self.tokenizer.train_from_iterator(iterator=iterator,
                                           vocab_size=self.vocab_size,
                                           min_frequency=self.min_frequency,
                                           special_tokens=self.special_tokens,
                                           show_progress=False)

    def save(self, output_dir: Text):
        """
        Save a trained tokenizer model to disk.

        Args:
            output_dir: Path to which to save the trained tokenizer.
        """

        # save_model does not attempt to create the destination, so create it
        path_utils.create_dir_if_not_exists(output_dir)

        self.tokenizer.save_model(directory=output_dir)

    def load_vocab(self, path_to_vocab: Text):
        """
        Re-instantiate the class tokenizer with output vocabulary / merges.

        Args:
            path_to_vocab: Path to vocab / merges files from a training run.
        """

        # inspect contents of output dir
        contents = path_utils.list_dir(path_to_vocab)

        try:
            vocab_file = next(f for f in contents if "vocab" in f)
        except StopIteration:
            vocab_file = None

        # update tokenizer params with vocab file name
        self.tokenizer_params.update({"vocab": vocab_file})

        # merges are only needed for BPE Tokenizers
        if "bpe" in self.tokenizer_name:
            try:
                merges_file = next(f for f in contents if "merge" in f)
            except StopIteration:
                merges_file = None

            self.tokenizer_params.update({"merges": merges_file})

        # reconstruct tokenizer object
        self.tokenizer = tokenizer_map.get(
            self.tokenizer_name)(**self.tokenizer_params)
        self.tokenizer.enable_padding(length=self.sentence_length)
        self.tokenizer.enable_truncation(max_length=self.sentence_length)

    def encode(self, sequence: Text, output_format: Text = "tf_example"):
        """
        Encode a sentence with the tokenizer in this class and output it into
        the specified output format. Supported return types are "tf_tensors",
        returning a dictionary of tensorflow tensor objects, "dict", returning
        a dictionary of lists, and "tf_example", returning a tf.train.Example
        proto file.

        Args:
            sequence: String, sentence to encode with the trained tokenizer.
            output_format: String specifying output format. Can be either
             `tf_tensors`, `dict` or `tf_example`.

        Returns:
            A representation of the sentence in IDs based on the vocabulary
             of the tokenizer.
        """
        eligible_formats = ["tf_tensors", "dict", "tf_example"]

        if output_format not in eligible_formats:
            raise ValueError("Unrecognized format {0}. Output format must be "
                             "one of: {1}".format(output_format,
                                                  ", ".join(eligible_formats)
                                                  ))

        encoded = self.tokenizer.encode(sequence=sequence)

        if output_format == "tf_example":
            feature = {"input_ids": tf.train.Feature(
                int64_list=tf.train.Int64List(value=encoded.ids)),
                "attention_mask": tf.train.Feature(
                    int64_list=tf.train.Int64List(
                        value=encoded.attention_mask))}

            output = tf.train.Example(
                features=tf.train.Features(feature=feature))

        elif output_format == "dict":
            output = {"input_ids": encoded.ids,
                      "attention_mask": encoded.attention_mask}

        else:
            # output format is tf_tensors
            output = {"input_ids": tf.convert_to_tensor(encoded.ids,
                                                        dtype=tf.int32),
                      "attention_mask": tf.convert_to_tensor(
                          encoded.attention_mask,
                          dtype=tf.int32)}
        return output



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import List, Text

import numpy as np
import tensorflow as tf
from joblib import dump
from sklearn import svm
from sklearn.metrics import classification_report, confusion_matrix

from zenml.steps.trainer import BaseTrainerStep
from zenml.steps.trainer import utils as trainer_utils
from zenml.utils import naming_utils
from zenml.utils import path_utils
from zenml.utils.post_training.post_training_utils import \
    convert_raw_dataset_to_pandas


class MyScikitTrainer(BaseTrainerStep):
    """
    Scikit trainer to train a scikit-learn based classifier model.
    """

    def __init__(self, c: float = 1.0, kernel: Text = 'rbf', **kwargs):
        self.c = c
        self.kernel = kernel

        super(MyScikitTrainer, self).__init__(
            c=self.c,
            kernel=self.kernel,
            **kwargs
        )

    def run_fn(self):
        train_split_patterns = [self.input_patterns[split] for split in
                                self.split_mapping[trainer_utils.TRAIN_SPLITS]]
        x_train, y_train = self.input_fn(train_split_patterns)

        eval_split_patterns = [self.input_patterns[split] for split in
                               self.split_mapping[trainer_utils.EVAL_SPLITS]]
        x_eval, y_eval = self.input_fn(eval_split_patterns)

        clf = svm.SVC(C=self.c, kernel=self.kernel)
        clf.fit(x_train, y_train)

        y_pred = clf.predict(x_eval)

        print("*********EVALUATION START******************")
        print(confusion_matrix(y_eval, y_pred))
        print(classification_report(y_eval, y_pred))
        print("*********EVALUATION DONE******************")

        # Save to serving_model_dir
        dump(clf, self.serving_model_dir)

        # for model agnostic evaluator
        for split in self.split_mapping[trainer_utils.TEST_SPLITS]:
            pattern = self.input_patterns[split]
            x_test, y_test = self.input_fn([pattern])
            test_results = self.test_fn(clf, x_test, y_test)
            trainer_utils.save_test_results(test_results,
                                            self.output_patterns[split])

    def input_fn(self,
                 file_pattern: List[Text]):
        """
        Load TFRecords on disk to pandas dataframe.

        Args:
            file_pattern: File pattern matching saved TFRecords on disk.

        Returns:
            dataset: tf.data.Dataset created out of the input files.
        """
        root_paths = [x.replace("*", "") for x in file_pattern]

        file_paths = []
        for root in root_paths:
            file_paths.extend(path_utils.list_dir(root))

        dataset = tf.data.TFRecordDataset(file_paths,
                                          compression_type='GZIP')
        df = convert_raw_dataset_to_pandas(dataset,
                                           self.schema,
                                           100000)

        # Separate labels
        X = df[[x for x in df.columns if
                naming_utils.check_if_transformed_feature(x)]]
        y = df[[x for x in df.columns if
                naming_utils.check_if_transformed_label(x)]]
        return X, y

    def test_fn(self, model, x_eval, y_eval):
        label_name = y_eval.columns[0]
        y_pred = model.predict(x_eval)

        out_dict = {label_name: y_eval.values,
                    naming_utils.output_name(label_name): y_pred}

        out_dict.update({k: np.array(v) for k, v
                         in x_eval.to_dict(orient='list').items()})
        return out_dict



from typing import Dict, Optional, Text, List

import apache_beam as beam
import tensorflow_model_analysis as tfma
from tensorflow_model_analysis import config
from tensorflow_model_analysis import constants
from tensorflow_model_analysis import model_util
from tensorflow_model_analysis import types
from tensorflow_model_analysis.extractors import extractor
from tfx_bsl.tfxio import tensor_adapter

BATCHED_PREDICT_EXTRACTOR_STAGE_NAME = 'ExtractBatchPredictions'


def custom_extractors(eval_config,
                      eval_shared_model,
                      tensor_adapter_config
                      ) -> List[tfma.extractors.Extractor]:
    return tfma.default_extractors(
        eval_config=eval_config,
        eval_shared_model=eval_shared_model,
        tensor_adapter_config=tensor_adapter_config,
        custom_predict_extractor=BatchedPredictExtractor(eval_config,
                                                         eval_shared_model,
                                                         tensor_adapter_config
                                                         ))


def BatchedPredictExtractor(
        eval_config: config.EvalConfig,
        eval_shared_model: types.MaybeMultipleEvalSharedModels,
        tensor_adapter_config: Optional[
            tensor_adapter.TensorAdapterConfig] = None,
) -> extractor.Extractor:
    eval_shared_models = model_util.verify_and_update_eval_shared_models(
        eval_shared_model)

    return extractor.Extractor(
        stage_name=BATCHED_PREDICT_EXTRACTOR_STAGE_NAME,
        ptransform=_ExtractBatchedPredictions(
            eval_config=eval_config,
            eval_shared_models={m.model_name: m for m in eval_shared_models},
            tensor_adapter_config=tensor_adapter_config))


@beam.ptransform_fn
@beam.typehints.with_input_types(types.Extracts)
@beam.typehints.with_output_types(types.Extracts)
def _ExtractBatchedPredictions(
        extracts: beam.pvalue.PCollection,
        eval_config: config.EvalConfig,
        eval_shared_models: Dict[Text, types.EvalSharedModel],
        tensor_adapter_config: Optional[
            tensor_adapter.TensorAdapterConfig] = None,
) -> beam.pvalue.PCollection:
    signature_names = {}
    for spec in eval_config.model_specs:
        model_name = '' if len(eval_config.model_specs) == 1 else spec.name
        signature_names[model_name] = [spec.signature_name]

    return (extracts
            | 'Predict' >> beam.ParDo(
                model_util.ModelSignaturesDoFn(
                    eval_config=eval_config,
                    eval_shared_models=eval_shared_models,
                    signature_names={
                        constants.PREDICTIONS_KEY: signature_names},
                    prefer_dict_outputs=True,
                    tensor_adapter_config=tensor_adapter_config)))



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import Text, List

import apache_beam as beam
import tensorflow as tf

from zenml.components.bulk_inferrer.constants import PREDICTIONS
from zenml.steps import BaseStep


class BaseInferrer(BaseStep):
    """
    Base inferrer class. This step is responsible for inference (batch).
    """

    def __init__(self,
                 labels: List[Text],
                 **kwargs):
        """
        Base Inferrer constructor.

        Args:
            model_uri: URI for a model, usually generated by
            TrainingPipeline and retrieved by
            `training_pipeline.get_model_uri()`.
        """
        self.labels = labels
        self.output_uri = None
        super(BaseInferrer, self).__init__(
            labels=labels,
            **kwargs,
        )

    def get_labels(self):
        return self.labels

    def set_output_uri(self, output_uri):
        self.output_uri = output_uri

    def write_inference_results(self):
        return beam.io.WriteToTFRecord(
            os.path.join(self.output_uri, PREDICTIONS),
            file_name_suffix='.gz',
            coder=beam.coders.ProtoCoder(tf.train.Example))



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of the DataFlow Processing Backend"""

import os
import time
from typing import Optional, List, Text

from zenml.backends.processing import ProcessingBaseBackend
from zenml.utils import requirement_utils
from zenml.constants import ZENML_DATAFLOW_IMAGE_NAME
from zenml.logger import get_logger

requirement_utils.check_integration(requirement_utils.GCP_INTEGRATION)

logger = get_logger(__name__)


class ProcessingDataFlowBackend(ProcessingBaseBackend):
    """
    Use this to run a ZenML pipeline on Google Dataflow.

    This backend utilizes the beam v2 runner to run a custom docker image on
    the Dataflow job.
    """

    def __init__(
            self,
            project: Text,
            staging_location: Text = None,
            temp_location: Text = None,
            region: Text = 'europe-west1',
            job_name: Text = f'zen_{int(time.time())}',
            image: Text = ZENML_DATAFLOW_IMAGE_NAME,
            machine_type: Text = 'n1-standard-4',
            num_workers: int = 4,
            max_num_workers: int = 10,
            disk_size_gb: int = 50,
            autoscaling_algorithm: Text = 'THROUGHPUT_BASED',
            **kwargs):
        """
        Adding this Backend will cause all 'Beam'-supported Steps in the
        pipeline to run on Google Dataflow.

        Args:
            project: GCP project to launch dataflow job.
            region: GCP region to launch dataflow job.
            job_name: Name of dataflow job.
            image: Docker Image to use. Must inherit from the beam base image.
            machine_type: Type of machine to run workload.
            num_workers: Number of workers on that machine.
            max_num_workers: Max number of workers in the workload.
            disk_size_gb: Disk size per worker.
            autoscaling_algorithm: Autoscaling algorithm to use.
            **kwargs:
        """
        self.project = project
        self.region = region
        self.job_name = job_name
        self.machine_type = machine_type
        self.num_workers = num_workers
        self.max_num_workers = max_num_workers
        self.disk_size_gb = disk_size_gb
        self.autoscaling_algorithm = autoscaling_algorithm
        self.image = image
        self.staging_location = staging_location
        self.temp_location = temp_location

        # staging and temp can be the same if one is specified
        if staging_location is not None and temp_location is None:
            self.temp_location = self.staging_location

        super().__init__(
            project=project,
            staging_location=staging_location,
            temp_location=temp_location,
            region=region,
            job_name=job_name,
            image=image,
            machine_type=machine_type,
            num_workers=num_workers,
            max_num_workers=max_num_workers,
            disk_size_gb=disk_size_gb,
            autoscaling_algorithm=autoscaling_algorithm,
            **kwargs
        )

    def get_beam_args(self,
                      pipeline_name: Text = None,
                      pipeline_root: Text = None) -> \
            Optional[List[Text]]:
        if self.temp_location is None:
            self.temp_location = os.path.join(
                pipeline_root, 'tmp', pipeline_name)
        if self.staging_location is None:
            self.staging_location = os.path.join(
                pipeline_root, 'staging', pipeline_name)

        return [
            '--runner=dataflow',
            '--project=' + self.project,
            '--temp_location=' + self.temp_location,
            '--staging_location=' + self.staging_location,
            '--region=' + self.region,
            # '--job_name=' + self.job_name,
            '--num_workers=' + str(self.num_workers),
            '--max_num_workers=' + str(self.max_num_workers),

            # Temporary overrides of defaults.
            '--disk_size_gb=' + str(self.disk_size_gb),
            '--experiments=shuffle_mode=auto',
            '--machine_type=' + self.machine_type,

            # Using docker
            '--experiment=use_runner_v2',
            f'--worker_harness_container_image={self.image}'
        ]



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import base64
import json
import os
from textwrap import dedent
from typing import Text

import click
import nbformat as nbf
import pandas as pd
import panel
import panel as pn
import tensorflow as tf
import tensorflow_data_validation as tfdv
import tensorflow_datasets as tfds
from tensorflow_metadata.proto.v0 import statistics_pb2
from tensorflow_transform.tf_metadata import schema_utils
from tfx.utils import io_utils

from zenml.constants import APP_NAME, EVALUATION_NOTEBOOK
from zenml.enums import GDPComponent
from zenml.logger import get_logger
from zenml.utils.path_utils import read_file_contents

logger = get_logger(__name__)


def get_statistics_html(stats_dict):
    """
    Args:
        stats_dict:
    """
    from tensorflow_metadata.proto.v0 import statistics_pb2

    combined_statistics = statistics_pb2.DatasetFeatureStatisticsList()

    for split, stats in stats_dict.items():
        stats_copy = combined_statistics.datasets.add()
        stats_copy.MergeFrom(stats.datasets[0])
        stats_copy.name = split

    protostr = base64.b64encode(
        combined_statistics.SerializeToString()).decode('utf-8')

    # assumes stats.html in the same folder
    template = \
        os.path.join(os.path.abspath(os.path.dirname(__file__)), 'stats.html')
    html_template = read_file_contents(template)

    # pylint: enable=line-too-long
    html = html_template.replace('protostr', protostr)
    return html


def get_statistics_dataset_dict(stats_uri: Text):
    """Get DatasetFeatureStatisticsList from stats URI"""
    result = {}
    for split in os.listdir(stats_uri):
        stats_path = os.path.join(stats_uri, split, 'stats_tfrecord')
        serialized_stats = next(
            tf.compat.v1.io.tf_record_iterator(stats_path))
        stats = statistics_pb2.DatasetFeatureStatisticsList()
        stats.ParseFromString(serialized_stats)
        dataset_list = statistics_pb2.DatasetFeatureStatisticsList()
        for i, d in enumerate(stats.datasets):
            d.name = split
            dataset_list.datasets.append(d)
        result[split] = dataset_list
    return result


def view_statistics(artifact_uri, magic: bool = False, port: int = 0):
    """
    View statistics in HTML.

    Args:
        artifact_uri (Text):
        magic (bool):
        port (int): Port at which to launch the visualization.
    """
    stats_dict = get_statistics_dataset_dict(artifact_uri)
    h = get_statistics_html(stats_dict)

    if magic:
        import sys
        if 'ipykernel' not in sys.modules:
            raise EnvironmentError(
                'The magic functions are only usable '
                'in a Jupyter notebook.')
        from IPython.core.display import display, HTML
        display(HTML(h))
    else:
        pn.serve(panels=pn.pane.HTML(h, width=1200), port=port, show=True)


def detect_anomalies(stats_uri: Text, schema_uri: Text, split_name: Text):
    schema = get_schema_proto(schema_uri)
    stats = get_statistics_dataset_dict(stats_uri)
    if split_name not in stats:
        raise Exception(f'{split_name} split not present!')
    anomalies = tfdv.validate_statistics(stats[split_name], schema)
    tfdv.display_anomalies(anomalies)


def view_schema(uri: Text):
    """
    View schema .

    Args:
        uri: URI to schema.
    """
    schema = get_schema_proto(uri)

    # TODO: [LOW] Replace with our own function
    tfdv.display_schema(schema=schema)


def get_schema_proto(artifact_uri: Text):
    schema_path = os.path.join(artifact_uri, 'schema.pbtxt')
    schema = io_utils.SchemaReader().read(schema_path)
    return schema


def get_feature_spec_from_schema(schema_uri):
    """
    Get schema artifact from pipeline

    Args:
        schema_uri: path to schema
    """
    schema_proto = get_schema_proto(schema_uri)
    spec = schema_utils.schema_as_feature_spec(schema_proto).feature_spec
    return spec


def get_parsed_dataset(dataset, spec):
    """
    Takes tf.data.Dataset and parses it based on spec

    Args:
        dataset: a tf.data.Dataset object
        spec: the spec to parse from
    """

    def parse(raw_record):
        return tf.io.parse_example(raw_record, spec)

    dataset = dataset.map(parse)
    return dataset


def convert_data_to_numpy(dataset, sample_size):
    """
    Takes tf.data.Dataset and converts to numpy array.

    Args:
        dataset: a tf.data.Dataset object
        sample_size: number of rows to limit to
    """

    # TODO: [MEDIUM] Check if this conversion to dense tensor makes sense
    def convert_if_sparse(element):
        return {k: tf.sparse.to_dense(tf.sparse.reorder(v))
        if type(v) == tf.sparse.SparseTensor else v
                for k, v in element.items()}

    dataset = dataset.map(convert_if_sparse)

    np_dataset = tfds.as_numpy(dataset)
    data = []
    for i, d in enumerate(np_dataset):
        if i == sample_size:
            break

        # usually v[0] has the data, but sometimes its an empty list
        new_row = {k: v[0] if len(v) != 0 else None for k, v in d.items()}

        # convert the bytes to strings for easier use
        new_row = {k: v if type(v) != bytes else v.decode()
                   for k, v in new_row.items()}
        data.append(new_row)
    return data


def convert_raw_dataset_to_pandas(dataset, spec, sample_size):
    """
    Takes tf.data.Dataset and converts to a Pandas DataFrame.

    Args:
        dataset: a tf.data.Dataset object
        spec: the spec to parse from
        sample_size: number of rows to limit to
    """
    logger.info('Converting dataset to Pandas DataFrame..')
    dataset = get_parsed_dataset(dataset, spec)
    data = convert_data_to_numpy(dataset, sample_size)
    return pd.DataFrame(data)


def create_new_cell(contents):
    """
    Creates new cell in jupyter notebook.

    Args:
        contents: contents of cell.
    """
    import sys
    from IPython.core.getipython import get_ipython

    if 'ipykernel' not in sys.modules:
        raise EnvironmentError('The magic functions are only usable in a '
                               'Jupyter notebook.')

    shell = get_ipython()

    payload = dict(
        source='set_next_input',
        text=contents,
        replace=False,
    )
    shell.payload_manager.write_payload(payload, single=False)


def get_tensorboard_block(log_dir):
    block = '''\
    import os

    model_path = '"{evaluation}"'
    logdir = os.path.join(model_path, 'logs')
    %load_ext tensorboard
    '''.format(evaluation=log_dir)
    block = dedent(block)
    block += '%tensorboard --logdir {logdir}'
    return block


def get_eval_block(eval_dir):
    block = '''\
    # If the visualization does not appear after running this block, please 
    # run the same cell again

    import tensorflow_model_analysis as tfma

    evaluation_path = '{evaluation}'
    evaluation = tfma.load_eval_result(output_path=evaluation_path)

    # find slicing metrics
    slicing_columns = set([x.feature_keys[0] for x in 
    evaluation.config.slicing_specs if len(x.feature_keys) == 1])
    print("")
    print("Available slicing columns: ")
    print(slicing_columns)

    tfma.view.render_slicing_metrics(evaluation)

    # in order to view sliced results, pass in the `slicing_column` parameter:
    # tfma.view.render_slicing_metrics(evaluation, slicing_column='col_name')
    '''.format(evaluation=eval_dir)
    return dedent(block)[:-1]


def evaluate_single_pipeline(
        pipeline,
        trainer_component_name: Text = None,
        evaluator_component_name: Text = None,
        magic: bool = False,
        port: int = 0):
    """
    Args:
        pipeline: A ZenML pipeline
        trainer_component_name: name of trainer component.
        evaluator_component_name: name of evaluator component.
        magic: Whether to create an in-place cell in a jupyter env or not.
        port: At which port to create the jupyter notebook.
    """
    # Default to the standard names
    trainer_component_name = trainer_component_name \
        if trainer_component_name else GDPComponent.Trainer.name
    evaluator_component_name = evaluator_component_name \
        if evaluator_component_name else GDPComponent.Evaluator.name

    trainer_path = pipeline.get_artifacts_uri_by_component(
        trainer_component_name)[0]
    eval_path = pipeline.get_artifacts_uri_by_component(
        evaluator_component_name)[0]

    # Patch to make it work locally
    with open(os.path.join(eval_path, 'eval_config.json'), 'r') as f:
        eval_config = json.load(f)
    eval_config['modelLocations'][''] = eval_path
    with open(os.path.join(eval_path, 'eval_config.json'), 'w') as f:
        json.dump(eval_config, f)

    if magic:
        tensorboard_block = get_tensorboard_block(trainer_path)
        eval_block = get_eval_block(eval_path)

        create_new_cell(eval_block)
        create_new_cell(tensorboard_block)
    else:
        nb = nbf.v4.new_notebook()
        nb['cells'] = [
            nbf.v4.new_code_cell(get_tensorboard_block(trainer_path)),
            nbf.v4.new_code_cell(get_eval_block(eval_path))]

        config_folder = click.get_app_dir(APP_NAME)

        if not (os.path.exists(config_folder) and os.path.isdir(
                config_folder)):
            os.makedirs(config_folder)

        final_out_path = os.path.join(config_folder, EVALUATION_NOTEBOOK)
        s = nbf.writes(nb)
        if isinstance(s, bytes):
            s = s.decode('utf8')

        with open(final_out_path, 'w') as f:
            f.write(s)

        if port == 0:
            os.system('jupyter notebook "{}"'.format(final_out_path))
        else:
            os.system(f'jupyter notebook {final_out_path} --port {port}')


def launch_compare_tool(port: int = 0, datasource=None):
    """Launches `compare` tool for comparing multiple training pipelines.

    Args:
        port: Port to launch application on.
        datasource (BaseDatasource): object of type BaseDatasource, to 
        filter only pipelines using that particular datasource.
    """
    from zenml.utils.post_training.compare import generate_interface
    panel.serve(generate_interface(datasource), port=port)



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Text, List

import apache_beam as beam

from zenml.steps.inferrer import BaseInferrer


class FileWriterInferrer(BaseInferrer):
    """
    File Writer Inferrer step, writing inference results to a text file.
    """

    def __init__(self,
                 labels: List[Text],
                 output_path: Text,
                 output_ext: Text = ".csv",
                 num_shards: int = 0,
                 **kwargs):
        """
        Base Tensorflow Inferrer constructor.

        Args:
            name: Outward-facing name of the pipeline.
        """
        self.labels = labels
        self.output_path = output_path
        self.output_ext = output_ext
        self.num_shards = num_shards

        super(FileWriterInferrer, self).__init__(
            labels=labels,
            **kwargs,
        )

    def get_destination(self):
        return beam.io.WriteToText(
            self.output_path,
            file_name_suffix=self.output_ext,
            append_trailing_newlines=True,
            num_shards=self.num_shards,
        )



"""

"""
import os

from zenml.backends.orchestrator import OrchestratorAWSBackend
from zenml.datasources import CSVDatasource
from zenml.exceptions import AlreadyExistsException
from zenml.metadata import MySQLMetadataStore
from zenml.pipelines import TrainingPipeline
from zenml.repo import ArtifactStore, Repository
from zenml.steps.evaluator import TFMAEvaluator
from zenml.steps.preprocesser import StandardPreprocesser
from zenml.steps.split import RandomSplit
from zenml.steps.trainer import TFFeedForwardTrainer
from zenml.utils.naming_utils import transformed_label_name

# Get the configuration for the artifact store and the metadata store which
# should be accessible from the VM
S3_BUCKET = os.getenv('S3_BUCKET')
IAM_ROLE = os.getenv('IAM_ROLE')

assert S3_BUCKET
assert IAM_ROLE

MYSQL_DB = os.getenv('MYSQL_DB')
MYSQL_USER = os.getenv('MYSQL_USER')
MYSQL_PWD = os.getenv('MYSQL_PWD')
MYSQL_HOST = os.getenv('MYSQL_HOST')
MYSQL_PORT = os.getenv('MYSQL_PORT', '3306')

assert MYSQL_DB
assert MYSQL_USER
assert MYSQL_PWD
assert MYSQL_HOST
assert MYSQL_PORT

# Define the training pipeline
training_pipeline = TrainingPipeline()

# Add a datasource. This will automatically track and version it.
try:
    ds = CSVDatasource(name='Pima Indians Diabetes AWS',
                       path='s3://zenml-quickstart/diabetes.csv')
except AlreadyExistsException:
    ds = Repository.get_instance().get_datasource_by_name(
        'Pima Indians Diabetes')
training_pipeline.add_datasource(ds)

# Add a split
training_pipeline.add_split(RandomSplit(
    split_map={'train': 0.7, 'eval': 0.2, 'test': 0.1}))

# Add a preprocessing unit
training_pipeline.add_preprocesser(
    StandardPreprocesser(
        features=['times_pregnant', 'pgc', 'dbp', 'tst', 'insulin', 'bmi',
                  'pedigree', 'age'],
        labels=['has_diabetes'],
        overwrite={'has_diabetes': {
            'transform': [{'method': 'no_transform', 'parameters': {}}]}}
    ))

# Add a trainer
training_pipeline.add_trainer(TFFeedForwardTrainer(
    loss='binary_crossentropy',
    last_activation='sigmoid',
    output_units=1,
    metrics=['accuracy'],
    epochs=20))

# Add an evaluator
training_pipeline.add_evaluator(
    TFMAEvaluator(slices=[['has_diabetes']],
                  metrics={transformed_label_name('has_diabetes'): [
                      'binary_crossentropy',
                      'binary_accuracy']}))

# Define the metadata store
metadata_store = MySQLMetadataStore(
    host=MYSQL_HOST,
    port=int(MYSQL_PORT),
    database=MYSQL_DB,
    username=MYSQL_USER,
    password=MYSQL_PWD,
)

# Define the artifact store
artifact_store = ArtifactStore(
    os.path.join(S3_BUCKET, 'aws_orchestrated/artifact_store'))

# Define the orchestrator backend
orchestrator_backend = OrchestratorAWSBackend(
    iam_role=IAM_ROLE
)

# Run the pipeline
training_pipeline.run(
    backend=orchestrator_backend,
    metadata_store=metadata_store,
    artifact_store=artifact_store,
)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Global config for the ZenML installation."""

import os
from uuid import uuid4
import click
from zenml import constants
from zenml.utils import path_utils
from zenml.utils import yaml_utils
from zenml.logger import get_logger

logger = get_logger(__name__)


# TODO: [MEDIUM] Optimize the reads and writes of this

class GlobalConfig(dict):
    """Class definition for the global config."""
    __instance__ = None

    def __init__(self, *args, **kwargs):
        """
        This is a Singleton class.

        Args:
            *args:
            **kwargs:
        """
        if GlobalConfig.__instance__ is None:
            self.path = os.path.join(
                GlobalConfig.get_config_dir(), 'info.json')
            # Create default global config if it does not exist.
            path_utils.create_dir_recursive_if_not_exists(
                GlobalConfig.get_config_dir())

            if path_utils.file_exists(self.path):
                # Load the config
                self.load()
            else:
                # Set up a default config
                # True by default but user is always asked
                self['analytics_opt_in'] = True

            # Create user ID will save the whole thing
            self.user_id = self.create_user_id()

            super(GlobalConfig, self).__init__(*args, **kwargs)
            GlobalConfig.__instance__ = self
        else:
            raise Exception("You cannot create another GlobalConfig class!")

    @staticmethod
    def get_instance():
        """ Static method to fetch the current instance."""
        if not GlobalConfig.__instance__:
            GlobalConfig()
        return GlobalConfig.__instance__

    @staticmethod
    def get_config_dir():
        """Gets config dir."""
        return click.get_app_dir(constants.APP_NAME)

    def create_user_id(self):
        """Creates user_id if it does not exist."""
        if 'user_id' not in self:
            self['user_id'] = str(uuid4())
            self.save()
        return self['user_id']

    def get_user_id(self):
        """Gets user_id from config. If not present, creates a new one."""
        if 'user_id' in self:
            return self['user_id']
        else:
            return self.create_user_id()

    def get_analytics_opt_in(self) -> bool:
        """Gets user_id from config. If not present, creates a new one."""
        if 'analytics_opt_in' in self:
            return self['analytics_opt_in']
        else:
            # assume False if this variable not found
            return False

    def set_analytics_opt_in(self, toggle: bool):
        """Set opt-in flag for analytics"""
        self.load()
        self['analytics_opt_in'] = toggle
        self.save()

    def load(self):
        """Load from YAML file."""
        # TODO: [LOW] Look at locking mechanism to avoid race conditions.
        try:
            tbu = yaml_utils.read_json(self.path)
        except Exception as e:
            logger.error("Unable to load YAML file due to error:\n" + str(e) + "\nUpdating it as empty value.")
            tbu = {}
        self.update(tbu)

    def save(self):
        """Save current config to YAML file"""
        try:
            yaml_utils.write_json(self.path, self)
        except Exception as e:
            logger.error('Unable to save as YAML File.')
            raise e



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


import tensorflow_transform as tft

from zenml.steps.preprocesser.standard_preprocesser.methods import \
    methods_nonseq_filling, methods_transform
from zenml.utils.preprocessing_utils import MethodDescriptions


class NonSeqFillingMethods(MethodDescriptions):
    MODES = {'min': (methods_nonseq_filling.min_f, []),
             'max': (methods_nonseq_filling.max_f, []),
             'mean': (methods_nonseq_filling.mean_f, []),
             'custom': (methods_nonseq_filling.custom_f, ['custom_value'])}


class TransformMethods(MethodDescriptions):
    MODES = {'scale_by_min_max': (tft.scale_by_min_max, ['min', 'max']),
             'scale_to_0_1': (tft.scale_to_0_1, []),
             'scale_to_z_score': (tft.scale_to_z_score, []),
             'tfidf': (tft.tfidf, ['vocab_size']),
             'compute_and_apply_vocabulary': (
                 tft.compute_and_apply_vocabulary, []),
             'ngrams': (tft.ngrams, ['ngram_range', 'separator']),
             'hash_strings': (tft.hash_strings, ['hash_buckets']),
             'pca': (tft.pca, ['output_dim', 'dtype']),
             'bucketize': (tft.bucketize, ['num_buckets']),
             'no_transform': (lambda x: x, []),
             'image_reshape': (methods_transform.decode_and_reshape_image, []),
             'load_binary_image': (methods_transform.load_binary_image,
                                   []),
             'one_hot_encode': (methods_transform.one_hot_encode, ['depth'])}



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.standards.standard_keys import StepKeys
from zenml.utils.source_utils import load_source_path_class


def run_fn(fn_args):
    fn_args_dict = fn_args.__dict__
    custom_config = fn_args_dict.pop('custom_config')

    # Get the user defined args
    args = custom_config.pop(StepKeys.ARGS)

    # We update users args first, because fn_args might have overlaps
    args.update(fn_args_dict)

    # Load the step, parameterize it and run it
    c = load_source_path_class(custom_config.pop(StepKeys.SOURCE))
    return c(**args).run_fn()
